package ru.ispras.telecomtesk.tlsv13.HandshakeMessages;

import java.lang.System;
import java.util.ArrayList;
import java.util.Random;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.RecordProtocol.ProtocolVersion;
import ru.ispras.telecomtesk.tlsv13.Mutation.MutateField;

public class Hello extends HandshakeMessage
{
	@MutateField
	public ProtocolVersion version;
	@MutateField
	public byte[] rnd = new byte[32]; // Random Hello 32 bytes
	@MutateField
	public byte[] sessionID; // 0..32 bytes
	@MutateField
	public short[] cipher_suites; // 2..2^16-2 bytes
	@MutateField
	public byte[] compression_methods; // 1..2^8-1  enum { null(0), (255) }
	@MutateField
	public ArrayList<TLSExtension> extensions; // 0..2^16-1 

	public Hello() {
		this(null, new short[1]);
	}

	public Hello(short[] cs) {
		this(null, cs);
	}

	public Hello(byte[] id, short[] cs) {
		this(new ProtocolVersion(), null, id, cs);
		// set random field
		Random rndObj = new Random(System.currentTimeMillis());
		rndObj.nextBytes(rnd);
		byte[] gmt_unix_time = ConvertTypes.hton((int)(System.currentTimeMillis() / 1000));
		System.arraycopy(gmt_unix_time, 0, rnd, 0, 4);
	}

	public Hello(byte[] r, byte[] id, short[] cs) {
		this(new ProtocolVersion(), r, id, cs);
	}

	public Hello(ProtocolVersion ver, byte[] r, byte[] id, short[] cs) {
		this(1, ver, r, id, cs, null, null);
		compression_methods = new byte[1];
		compression_methods[0] = 0; // null method
	}

	public Hello(ProtocolVersion ver, byte[] r, byte[] id, short[] cs, byte[] cm, ArrayList<TLSExtension> ext) {
		this(1, ver, r, id, cs, cm, ext);
	}

	public Hello(int type, ProtocolVersion ver, byte[] r, byte[] id, short[] cs, byte[] cm, ArrayList<TLSExtension> ext) {
		msg_type = (byte)type;
		version = ver;
		if(r != null) {
			int len = r.length >= 32 ? 32 : r.length;
			System.arraycopy(r, 0, rnd, 0, len);
		}
		if(id != null)
			sessionID = id;
		else
			sessionID = new byte[0];
		cipher_suites = cs;
		compression_methods = cm;
		if(ext != null)
			extensions = ext;
		else
			extensions = new ArrayList<TLSExtension>(); 
	}

	public int getGMTunixTimeFromRandomValue() {
		return ConvertTypes.ntoh(rnd);
	}
	public boolean isSessionIDEmpty() {
		return (sessionID == null) || (sessionID.length == 0);
	}
	public boolean compareSessionID(byte[] id) {
		boolean res = false;
		int i;
		DebugInfo.printDump(4,sessionID,"Hello::compareSessionID, Hello id:");
		DebugInfo.printDump(4,id,"compared id:");
		if((id != null) && (sessionID != null) && (sessionID.length > 0) && (sessionID.length == id.length)){
			for(i=0,res = true; i < sessionID.length; i++)
				if(sessionID[i] != id[i]) {
					res = false;
					break;
				}
		}
		DebugInfo.printMessage(4,"compareSessionID return " + res);
		return res;
	}
	
	public void addExtension(TLSExtension ext) {
		extensions.add(ext);
	}
	public void removeExtension(TLSExtension ext) {
		extensions.remove(ext);
	}
	public void clearExtensionList() {
		extensions.clear();
	}
	
	@Override
	public int getSize() {
		int size;
		size = 42 + sessionID.length;
		if(msg_type == 1) 
			size += 2*cipher_suites.length + compression_methods.length;
		if(TLSExtensionsList.getExtensionLength(extensions) > 0)
			size += 2 + TLSExtensionsList.getExtensionLength(extensions);
		return size;
	}
	
	public void setTypeClientHello() {
		msg_type = 1;
	}
	
	public void setTypeServerHello() {
		msg_type = 2;
	}
	
	public boolean checkSCSV() {
		boolean res = false;
		for(int i=0; i<cipher_suites.length; i++)
			if(TLSCipherSuite.valueOf(cipher_suites[i]) == TLSCipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV) {
				res = true;
				break;
			}
		return res;
	}

	public boolean checkCipherSuite(short cs) {
		boolean res = false;
		if(this.cipher_suites != null)
			for(short csCur : this.cipher_suites) {
				if(csCur == cs) {
					res = true;
					break;
				}
			}
		return res;
	}
	
/* 
 * (non-Javadoc)
 * @see ru.ispras.telecomtesk.tls.HandshakeMessages.HandshakeMessage#toByteArray()
 *
 * size of Hello in bytes:
 * type + length = 4
 * version = 2
 * random = 32
 * <field 'length' of 1 byte> + sessionID = 1 + sessionID.length
 * <field 'length' of 2 byte> + cipher_suites = 2 + (2 * cipher_suites.length) // = 2 for ServerHello
 * <field 'length' of 1 byte> + compression_methods = 1 + compression_methods.length // = 1 for ServerHello
 * <??? field 'length' of 2 byte> + extensions = 2 + ...
 * 
 * ClientHello:
 * 4 + 2 + 32 + 1 + sessionID.length + 2 + 2*cipher_suites.length + 1 + compression_methods.length + 2 + <extensions>
 * 
 * ServerHello:
 * 4 + 2 + 32 + 1 + sessionID.length + 2 + 1 + 2 + <extensions>
 */
	public int toByteArray() {
		int i = 0, j, res = 0;
		if(isLengthSet == false) setLength();
		// create byte array
		byte[] ba = new byte[getSize()];
		// write msg_type and full length
		byte[] ba_len = ConvertTypes.hton(length);
		ba[0] = msg_type;
		System.arraycopy(ba_len, 1, ba, 1, 3);
		i=4;
		ba[i++] = version.major;
		ba[i++] = version.minor;
		// write RandomHello rnd
		System.arraycopy(rnd, 0, ba, i, rnd.length);
		i+=rnd.length;
		// write sessionID
		ba[i++] = (byte) sessionID.length;
		System.arraycopy(sessionID, 0, ba, i, (byte)sessionID.length);
		i+= (byte)sessionID.length;
		if(msg_type != 2) {
			// write cipher_suites
			j = 2 * cipher_suites.length;
			ba[i++] = (byte)(j >>> 8);
			ba[i++] = (byte)j;
			for (j = 0; j < cipher_suites.length; j++) {
				ba[i++] = (byte)(cipher_suites[j] >>> 8);
				ba[i++] = (byte)cipher_suites[j];
			}
			// write compression_methods
			ba[i++] = (byte) compression_methods.length;
			System.arraycopy(compression_methods, 0, ba, i, (byte)compression_methods.length);
			i+= (byte)compression_methods.length;
		}
		else {
			// write cipher_suites
			ba[i++] = (byte)(cipher_suites[0] >>> 8);
			ba[i++] = (byte)cipher_suites[0];
			// write compression_methods
			ba[i++] = (byte) compression_methods[0];
		}
		//write extensions
		TLSExtensionsList.toByteArray(ba, i, extensions);
		
		net_msg = ba;
		return res;
	}

	public String toString() {
		String res;

		if(msg_type == 1) {
			res = "HandshakeMessage: type ClientHello, length " + length;
			if(net_msg != null)
				res += ", net_msg: " + net_msg.length;
		}
		else if(msg_type == 2) {		
                	int i;
			res = "HandshakeMessage: type ServerHello";
			if(isHelloRetryRequest()) 
				res += "[HelloRetry]";
			res += ", length " + length;
			if(net_msg != null)
				res += ", net_msg: " + net_msg.length;

			res += "\n      ProtocolVersion: " + version.major + "." + version.minor;
			res += "\n      cipher_suites: ";
			for(i=0;i<cipher_suites.length;i++) {
				res += TLSCipherSuite.valueOf(cipher_suites[i]) + ";";
			}

			res += "\n      compression_methods: ";

			for(i=0;i<compression_methods.length;i++) {
				res += compression_methods[i] + ";";
			}

		}
		else {		
			res = "HandshakeMessage: Hello with type " + msg_type + ", length " + length;
			if(net_msg != null)
				res += ", net_msg: " + net_msg.length;
		}

		return res;
	}

	public void print() {
		int i;
		DebugInfo.printMessage(1,"\nHandshake Hello Message:");
		DebugInfo.printMessage(1,"ProtocolVersion: " + version.major + " " + version.minor);
		if(isHelloRetryRequest())
			DebugInfo.printMessage(1,"It's HelloRetry Message");
		DebugInfo.printDump(1,rnd,"Random:");
		DebugInfo.printDump(1,sessionID,"session ID:");
		DebugInfo.printMessage(1,"cipher_suites, " + cipher_suites.length + ":");
		for(i=0;i<cipher_suites.length;i++) {
			DebugInfo.printMessage(1,"   " + TLSCipherSuite.toString(cipher_suites[i]));
		}
		DebugInfo.printMessage(1,"compression_methods, " + compression_methods.length + ":");
		System.out.print("   ");
		for(i=0;i<compression_methods.length;i++) {
			System.out.printf("%d ",compression_methods[i]);
		}
		TLSExtensionsList.print(this.extensions);
	}

	// the HelloRetryRequest message uses the same structure
	// as the ServerHello, but with Random set to the special value 
	public static boolean isHelloRetryRequest(byte[] rnd) {
		return ConvertTypes.compareArrays(rnd, TLS13SpecValues.randomHelloRetryRequest);
	}
	
	public boolean isHelloRetryRequest() {
		return Hello.isHelloRetryRequest(this.rnd);
	}
	
    // If negotiating TLS 1.2 or 1.1, TLS 1.3 servers MUST set the last 8 bytes
    // of their Random value to the bytes
	// For TLS 1.2 return 2, for TLS 1.1 return 1, otherwise return 0 
	public int checkTLSDowngradeProtection() {
		int res = 0 ,i;
		if((this.rnd != null) && (this.rnd.length == 32)) {
			for(i=0;i<7;i++) {
				if(this.rnd[24 + i] != TLS13SpecValues.randomDowngradeProtectionTLS12[i])
					break;
			}
			if(i == 7) {
				if(this.rnd[31] == TLS13SpecValues.randomDowngradeProtectionTLS12[7])
					res = 2;
				else if(this.rnd[31] == TLS13SpecValues.randomDowngradeProtectionTLS11[7])
					res = 1;
			}
		}
		return res;
	}
	
    public static HandshakeMessage parseByteArray(byte[] msg) {
    	HandshakeMessage hm = null;
    	int pos = 0, len, i;
    	boolean isHelloRetryMessage = false;

    	// 39 bytes: 4 (header) + 2 (ProtocolVersion) + 32 (Random) + 1 (SessionID length)
    	if(msg.length < 39) {
    		DebugInfo.printMessage(4,"===Error! Hello::parseByteArray: message header length failed");
    		return null;
		}
    	ProtocolVersion pv = new ProtocolVersion(msg[4],msg[5]);
    	byte[] rnd = new byte[32];
    	System.arraycopy(msg, 6, rnd, 0, 32);
    	
    	// is Hello Retry Message
    	isHelloRetryMessage = isHelloRetryRequest(rnd);

    	// parse SessionID
    	len = msg[38];
    	if(msg.length < 39 + len) {
    		DebugInfo.printMessage(4,"===Error! Hello::parseByteArray: SessionID failed");
    		return null;
		}
    	byte[] sid = new byte[len];
    	if(len > 0)
    		System.arraycopy(msg, 39, sid, 0, len);
    	pos = 39 + len;
    	
		short[] cipher_suites;
    	byte[] compression_methods;

    	// Client Hello
    	if(msg[0] != 2) { 
        	// parse CipherSuite
    		if(msg.length < pos + 2) {
    			DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ClientHello: CipherSuite length failed");
    			return null;
    		}
    		len = ConvertTypes.ntohs(msg,pos);
    		pos += 2;
    		if(msg.length < pos + len) {
    			DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ClientHello: CipherSuite failed");
    			return null;
    		}
    		if(len % 2 != 0) {
    			DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ClientHello: CipherSuite length must be even");
    			return null;
    		}
    		cipher_suites = new short[len/2];
    		for(i=0;i<cipher_suites.length;i++) {
    			cipher_suites[i] = ConvertTypes.ntohs(msg,pos);
    			pos += 2;
    		}
        	// parse CompressionMethod
        	if(msg.length < pos + 1) {
        		DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ClientHello: CompressionMethod length failed");
        		return null;
    		}
        	len = msg[pos++];
        	if(msg.length < pos + len) {
        		DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ClientHello: CompressionMethod failed");
        		return null;
    		}
        	compression_methods = new byte[len];
        	if(len > 0) {
        		System.arraycopy(msg, pos, compression_methods, 0, len);
        		pos += len;
        	}
    	}
    	// Server Hello
    	else {
        	// parse CipherSuite (2 bytes) and CompressionMethod (1 byte)
    		if(msg.length < pos + 3) {
    			DebugInfo.printMessage(4,"===Error! Hello::parseByteArray:ServerHello: CipherSuite length failed");
    			return null;
    		}
    		cipher_suites = new short[1];
   			cipher_suites[0] = ConvertTypes.ntohs(msg,pos);
   			pos += 2;
        	// parse CompressionMethod
        	compression_methods = new byte[1];
        	compression_methods[0] = msg[pos++];
    	}
    	
    	// parse Extensions
    	ArrayList<TLSExtension> extensions = TLSExtensionsList.parseTLSExtensionsList(msg, pos, isHelloRetryMessage);
   		len = 2 + TLSExtensionsList.getExtensionLength(extensions);
   		if(msg.length - pos > len) {
    		DebugInfo.printDumpWithLength(1, msg, pos+len, "===Error! Hello::parseByteArray: some bytes remain after Extensions");   			
   		}
   		
   		if(isHelloRetryMessage)
   			hm = new HelloRetryRequest(pv,sid,cipher_suites,compression_methods,extensions);
   		else
   			hm = new Hello(msg[0],pv,rnd,sid,cipher_suites,compression_methods,extensions);
   		
    	return hm;
    }
    	
}
