package ru.ispras.telecomtesk.tlsv13.HandshakeMessages;

import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.TLSExtension;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.TLSExtensionsList;

public class NewSessionTicket extends HandshakeMessage
{
    public int ticketLifetime; // uint32
    public int ticketAgeAdd;   // uint32
	private byte[] ticketNonce;    // <0..255>
	private byte[] ticket;    // <1..2^16-1>
	private ArrayList<TLSExtension> extensions; // <0..2^16-1> 
	
	public NewSessionTicket() {
		this(0,0,null, null, null);
	}

	public NewSessionTicket(byte[] ticketNonce, byte[] ticket) {
		this(0,0,ticketNonce, ticket, null);
	}

	public NewSessionTicket(int ticketLifetime, int ticketAgeAdd, byte[] ticketNonce, byte[] ticket) {
		this(ticketLifetime,ticketAgeAdd,ticketNonce, ticket, null);
	}

	public NewSessionTicket(int ticketLifetime, int ticketAgeAdd, byte[] ticketNonce, byte[] ticket, ArrayList<TLSExtension> extensions) {
		msg_type = 4;
		this.ticketLifetime = ticketLifetime;
		this.ticketAgeAdd = ticketAgeAdd;
		this.ticketNonce = ticketNonce != null ? ticketNonce : new byte[0];
		this.ticket = ticket != null ? ticket : new byte[0];
		this.extensions = extensions != null ? extensions : new ArrayList<TLSExtension>();
	}

	public byte[] getTicketNonce() {
		return ticketNonce;
	}

	public byte[] getTicket() {
		return ticket;
	}

	@Override
	public int getSize() {
		return (4 + 8 + 1 + ticketNonce.length + 2 + ticket.length + 2 + TLSExtensionsList.getExtensionLength(extensions));
	}

	public ArrayList<TLSExtension> getExtensions() {
		return extensions;
	}

	public void addExtension(TLSExtension ext) {
		extensions.add(ext);
	}
	
	public void removeExtension(TLSExtension ext) {
		extensions.remove(ext);
	}
	
	public void clearExtensionList() {
		extensions.clear();
	}

	@Override
	public String toString() {
		String res;

		res = "HandshakeMessage: type NewSessionTicket, length " + length;
		if(net_msg != null)
				res += ", net_msg: " + net_msg.length;

		return res;
	}

	@Override
	public void print() {
		DebugInfo.printMessage(1,"\nHandshake NewSessionTicket Message:");
		DebugInfo.printMessage(1,"ticketLifetime " + ticketLifetime + " ticketAgeAdd " + ticketAgeAdd);
		DebugInfo.printDump(1, ticketNonce, "ticketNonce:");
		DebugInfo.printDump(1, ticket, "ticket:");
		TLSExtensionsList.print(this.extensions);
	}

	@Override
	public int toByteArray() {
		if(isLengthSet == false) setLength();
		// create byte array
		byte[] ba = new byte[getSize()];
		// write msg_type and full length
		byte[] ba_len = ConvertTypes.hton(length);
		ba[0] = msg_type;
		System.arraycopy(ba_len, 1, ba, 1, 3);
		int i = 4;
		// ticketLifetime, ticketAgeAdd
		byte[] tlt = ConvertTypes.hton(ticketLifetime);
		System.arraycopy(tlt, 0, ba, i, tlt.length);
		i += tlt.length;
		byte[] ta = ConvertTypes.hton(ticketAgeAdd);
		System.arraycopy(ta, 0, ba, i, ta.length);
		i += ta.length;
		// ticketNonce and ticket
		ba[i++] = (byte)ticketNonce.length;
		System.arraycopy(ticketNonce, 0, ba, i, ticketNonce.length);
		i += ticketNonce.length;
		ba[i++] = (byte)(ticket.length >>> 8);
		ba[i++] = (byte)ticket.length;
		System.arraycopy(ticket, 0, ba, i, ticket.length);
		i += ticket.length;
		//write extensions
		TLSExtensionsList.toByteArray(ba, i, extensions);
		
		net_msg = ba;
		return 0;
	}
	
	public static HandshakeMessage parseByteArray(byte[] msg) {
    	int pos = 12, curlen;

    	if(msg.length <= pos) {
    		DebugInfo.printMessage(4,"===Error! NewSessionTicket::parseByteArray: CertificateRequestContext length failed");
    		return null;
		}
        int ticketLifetime = ConvertTypes.ntoh(msg,4); // uint32
        int ticketAgeAdd = ConvertTypes.ntoh(msg,8);   // uint32
        
    	// parse ticketNonce and ticket
    	curlen = msg[pos++];
    	if(msg.length - pos < curlen) {
    		DebugInfo.printMessage(4,"===Error! NewSessionTicket::parseByteArray: ticketNonce length wrong");
    		return null;
		}
    	byte[] ticketNonce = new byte[curlen];
		System.arraycopy(msg, pos, ticketNonce, 0, curlen);
		pos += curlen;
    	
		curlen = ConvertTypes.ntohs(msg,pos);
		pos += 2;
    	if(msg.length - pos < curlen) {
    		DebugInfo.printMessage(4,"===Error! NewSessionTicket::parseByteArray: ticket length wrong");
    		return null;
		}
    	byte[] ticket = new byte[curlen];
		System.arraycopy(msg, pos, ticket, 0, curlen);
		pos += curlen;
    	
    	// parse Extensions
    	ArrayList<TLSExtension> extensions = TLSExtensionsList.parseTLSExtensionsList(msg, pos);

		return new NewSessionTicket(ticketLifetime,ticketAgeAdd,ticketNonce,ticket,extensions);
	}
	    	
}
