package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import java.math.BigInteger;
import java.security.PublicKey;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECPoint;
import java.util.ArrayList;
import java.util.Arrays;

import javax.crypto.interfaces.DHPublicKey;
import javax.xml.bind.DatatypeConverter;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.NamedGroup;

public class KeyShareEntry
{
	private NamedGroup group; 
	public byte[] keyExchange; // <1..2^16-1> bytes

	public KeyShareEntry(NamedGroup group, byte[] keyExchange) {
		this.group = group; 
		this.keyExchange = keyExchange;
	}

	public KeyShareEntry(NamedGroup group, PublicKey pubKey) throws Exception {
		this.group = group;
		this.keyExchange = null;
		
		if(group.isDHE()) {
			byte[] DH_Yc = ((DHPublicKey)pubKey).getY().toByteArray();
			if(DH_Yc[0] == 0x00) 
			    DH_Yc = Arrays.copyOfRange(DH_Yc, 1, DH_Yc.length);
			this.keyExchange = DH_Yc;
		} else if(group.isECDHE()) {
	        ECPublicKey clientPublicKey = (ECPublicKey)pubKey; 
	        byte[] EC_Xc = clientPublicKey.getW().getAffineX().toByteArray();
	        byte[] EC_Yc = clientPublicKey.getW().getAffineY().toByteArray();
	        
	        int len = 1 + EC_Xc.length + EC_Yc.length;
	        int pos = 1;
	        if(EC_Xc.length % 2 != 0) 
	        	len--;
	        if(EC_Yc.length % 2 != 0) 
	        	len--;
	        
	        byte[] ec = new byte[len];
	        ec[0] = 4;
	        if(EC_Xc.length % 2 != 0) { 
				System.arraycopy(EC_Xc, 1, ec, pos, EC_Xc.length-1);
				pos += EC_Xc.length-1;
	        }
	        else {
				System.arraycopy(EC_Xc, 0, ec, pos, EC_Xc.length);
				pos += EC_Xc.length;
	        }
	        	
	        if(EC_Yc.length % 2 != 0) 
				System.arraycopy(EC_Yc, 1, ec, pos, EC_Yc.length-1);
	        else
				System.arraycopy(EC_Yc, 0, ec, pos, EC_Yc.length);
	        
			this.keyExchange = ec;
		}
	}

	public NamedGroup getGroup() {
		return group;
	}

	public int getSize() {
		if(keyExchange != null)
			return (4 + keyExchange.length);
		else
			return 4;
	}

	public static boolean checkExistKeyShare(ArrayList<KeyShareEntry> ksList, NamedGroup ng) {
		boolean res = false;
		if(ksList != null)
			for(KeyShareEntry curKS : ksList) {
				if((curKS != null) && (curKS.getGroup() == ng)) {
					res = true;
					break;
				}
			}
		return res;
	}

	public static boolean checkExistKeyShare(ArrayList<KeyShareEntry> ksList, KeyShareEntry ks) {
		boolean res = false;
		if(ks != null)
			res = checkExistKeyShare(ksList, ks.getGroup());
		return res;
	}

	public static KeyShareEntry getPairKeyShareEntry(ArrayList<KeyShareEntry> ksList, KeyShareEntry ks) {
		DebugInfo.printMessage(5,"KeyShareEntry::getPairKeyShareEntry");
		if((ksList != null) && (ks != null)) {
			for(int i = 0; i<ksList.size(); i++) {
				if(ksList.get(i).group == ks.group) { 
					DebugInfo.printMessage(5,"KeyShareEntry::getPairKeyShareEntry: ListSize index " + ksList.size() + " " + i);
					return ksList.get(i);
				}
			}
		}
		return null;
	}

	public static int getPairKeyShareIndex(ArrayList<KeyShareEntry> ksList, KeyShareEntry ks) {
		DebugInfo.printMessage(5,"KeyShareEntry::getPairKeyShareIndex");
		int res = -1;
		if((ksList != null) && (ks != null)) {
			for(int i = 0; i<ksList.size(); i++) {
				if(ksList.get(i).group == ks.group) {
					DebugInfo.printMessage(5,"KeyShareEntry::getPairKeyShareIndex: ListSize index " + ksList.size() + " " + i);
					return i;
				}
			}
		}
		return res;
	}
	
	public ECPoint getECPoint() {
		ECPoint ecP = null;
		byte[] ecX = null;
		byte[] ecY = null;

		if((keyExchange != null)&&(keyExchange.length > 1)) {
			int len = keyExchange.length - 1;
    		if(len % 2 != 0) {
    			DebugInfo.printMessage(4,"===Error! KeyShareEntry::getECPoint: keyExchange length-1 must be even");
    			return null;
    		}
			len /= 2;
			ecX = new byte[len];
			ecY = new byte[len];
			System.arraycopy(keyExchange, 1, ecX, 0, len);
			System.arraycopy(keyExchange, 1+len, ecY, 0, len);
			
			DebugInfo.printDump(5, ecX, "KeyShareEntry::getECPoint:ecX: ");
			DebugInfo.printDump(5, ecY, "KeyShareEntry::getECPoint:ecY: ");
			
		    String strX = DatatypeConverter.printHexBinary(ecX);
		    String strY = DatatypeConverter.printHexBinary(ecY);
			BigInteger rXb = new BigInteger(strX,16);
			BigInteger rYb = new BigInteger(strY,16);
			
//			BigInteger rXb = new BigInteger(ecX);
//			BigInteger rYb = new BigInteger(ecY);
			ecP = new ECPoint(rXb,rYb);
		}
		return ecP;
	}

	public byte[] toByteArray() {
		int i = 0;
		byte[] res = new byte[getSize()];
		int len = res.length - 4;
		int grp = group.getValue();
		res[i++] = (byte)(grp >>> 8);
		res[i++] = (byte)grp;
		res[i++] = (byte)(len >>> 8);
		res[i++] = (byte)len;
		if(keyExchange != null)
			System.arraycopy(keyExchange, 0, res, i, keyExchange.length);
		return res;
	}

	public int toByteArray(byte[] msg, int position) {
		int len = this.getSize();
		if((msg == null) || (position < 0) || (msg.length - position < len))
			return -1;

		int grp = group.getValue();
		len -= 4;
		msg[position++] = (byte)(grp >>> 8);
		msg[position++] = (byte)grp;
		msg[position++] = (byte)(len >>> 8);
		msg[position++] = (byte)len;
		if(keyExchange != null)
			System.arraycopy(keyExchange, 0, msg, position, keyExchange.length);
		position += keyExchange.length;

		return position;
	}

	public void print() {
		DebugInfo.printMessage(1,"  NamedGroup: " + group.toString());
		DebugInfo.printDump(1,keyExchange,"  keyExchange:");
	}

	public static KeyShareEntry parseByteArray(byte[] msg, int pos) {
		int len;
		int lengthExt = msg.length - pos;
    	if(lengthExt < 4) {
    		DebugInfo.printMessage(4,"===Error! KeyShareEntry::parseByteArray: length failed");
    		return null;
    	}

    	NamedGroup grp = NamedGroup.valueOf(ConvertTypes.ntohs(msg, pos));
    	pos += 2;
    	len = ConvertTypes.ntohs(msg,pos);
		pos += 2;

    	if(lengthExt < 4 + len) {
    		DebugInfo.printMessage(4,"===Error! KeyShareServerHello::parseByteArray: Extension length wrong");
    		return null;
    	}
    	
    	byte[] keydata = new byte[len];
		System.arraycopy(msg, pos, keydata, 0, len);
		
		return new KeyShareEntry(grp,keydata);
	}
}
