package ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions;

import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;

public class PreSharedKeyClientHello extends TLSExtension
{
	public ArrayList<PskIdentity> identities; // <7..2^16-1> bytes 
	public ArrayList<byte[]> binders; // <33..2^16-1> bytes 

	public PreSharedKeyClientHello() {
		this.type = 41;
		this.identities = new ArrayList<PskIdentity>();
		this.binders = new ArrayList<byte[]>();
	}

	public PreSharedKeyClientHello(ArrayList<PskIdentity> identities, ArrayList<byte[]> binders) {
		this.type = 41;
		this.identities = identities;
		this.binders = binders;
	}

	public PreSharedKeyClientHello(PskIdentity identity, byte[] pskBinder) {
		this();
		this.identities.add(identity);
		this.binders.add(pskBinder);
	}

	public PreSharedKeyClientHello(int length , ArrayList<PskIdentity> identities, ArrayList<byte[]> binders) {
		this(identities, binders);
		this.setLength(length);
	}

	public int getIdentitiesSize() {
		int res = 0;  
		for (PskIdentity identity : identities) {
			res += identity.getSize();
		}
		return res;
	}

	public int getBindersSize() {
		int res = 0;  
		for (byte[] pskBinder : binders) {
			if(pskBinder != null) {
				res += (1 + pskBinder.length);
			}
		}
		return res;
	}

	@Override
	public int getDataSize() {
		return (4 + getIdentitiesSize() + getBindersSize());  
	}

	@Override
	public void toByteArrayData(byte[] msg, int position) {
		int dataLen = getIdentitiesSize();
		msg[position++] = (byte)(dataLen >>> 8);
		msg[position++] = (byte)dataLen;
		if(identities != null) {
			for (int j=0; j < identities.size(); j++) {
				int curPos = (identities.get(j)).toByteArray(msg,position);
				if(curPos >= 0)
					position = curPos;
				else
					DebugInfo.printMessage(1,"!!! Error: PreSharedKeyClientHello toByteArray return null, identity[" + j + "]");
			}
		}

		dataLen = getBindersSize();
		msg[position++] = (byte)(dataLen >>> 8);
		msg[position++]   = (byte)dataLen;
		if(binders != null) {
			for (int j=0; j < binders.size(); j++) {
				byte[] curBinder = binders.get(j);
				if(curBinder != null) {
					msg[position++] = (byte)curBinder.length;
					System.arraycopy(curBinder, 0, msg, position, curBinder.length);
					position += curBinder.length;
				}
				else
					DebugInfo.printMessage(1,"!!! Error: PreSharedKeyClientHello toByteArray return null, binder[" + j + "]");
			}
		}

	}

	@Override
	public void printData() {
		DebugInfo.printMessage(1,"  PreSharedKeyClientHello");
		DebugInfo.printMessage(1,"     PskIdentities");
		for (PskIdentity identity : identities) {
			identity.print();
		}
		DebugInfo.printMessage(1,"     PskBinders");
		for (byte[] pskBinder : binders) {
			if(pskBinder != null) {
				DebugInfo.printDump(1,pskBinder,"       pskBinder:");
			}
		}
	}

	public static TLSExtension parseByteArray(int lengthExt, byte[] msg, int pos) {
		int len;
		int lengthFieldSize = 2;
		
		// parse identities
    	if(lengthExt < lengthFieldSize) {
    		DebugInfo.printMessage(4,"===Error! PreSharedKeyClientHello::parseByteArray: identities length failed");
    		return null;
    	}
		len = ConvertTypes.ntohs(msg,pos);
		pos += lengthFieldSize;

    	if(lengthExt < lengthFieldSize + len) {
    		DebugInfo.printMessage(4,"===Error! PreSharedKeyClientHello::parseByteArray: identities length wrong");
    		return null;
    	}
		lengthExt -= (lengthFieldSize + len); // for binders parse
    	
    	ArrayList<PskIdentity> identities = new ArrayList<PskIdentity>();
    	int curLen;

		while(len > 0) {
			PskIdentity curId = PskIdentity.parseByteArray(msg,pos);
			if(curId == null) 
				return null;
			identities.add(curId);
			curLen = curId.getSize();
    		pos += curLen;
    		len -= curLen;
		}

		// parse binders
    	if(lengthExt < lengthFieldSize) {
    		DebugInfo.printMessage(4,"===Error! PreSharedKeyClientHello::parseByteArray: binders length failed");
    		return null;
    	}
		len = ConvertTypes.ntohs(msg,pos);
		pos += lengthFieldSize;

    	if(lengthExt < lengthFieldSize + len) {
    		DebugInfo.printMessage(4,"===Error! PreSharedKeyClientHello::parseByteArray: binders length wrong");
    		return null;
    	}
    	
    	ArrayList<byte[]> binders = new ArrayList<byte[]>();

		while(len > 0) {
			curLen = msg[pos++];  // binder length
			if(curLen < 0) {
	    		DebugInfo.printMessage(4,"===Error! PreSharedKeyClientHello::parseByteArray: binder length < 0");
				return null;
			}
			byte[] curBinder = new byte[curLen];
			System.arraycopy(msg, pos, curBinder, 0, curBinder.length);
			binders.add(curBinder);
    		pos += curLen;
    		len -= (1 + curLen);
		}

		
		return new PreSharedKeyClientHello(lengthExt, identities, binders);
	}
}
