package ru.ispras.telecomtesk.tlsv13.RecordProtocol;

import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.HandshakeMessage;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;
import ru.ispras.telecomtesk.tlsv13.TLSState.TLSConnection;

public class RecordLayer 
{
	public TLSContentType tls_type;
	public TLSContentType opaque_tls_type;
	public ProtocolVersion version;
	protected short length;
	public byte[][] content;
	protected boolean isLengthSet = false;
	public boolean incompleteArrayMessage = false;
	public int error = 0;
	public ArrayList msglist;

	public static int MAX_LENGTH_TLSInnerPlaintext = 16383;
	
	public RecordLayer() {
		this(TLSContentType.HANDSHAKE);
	}
	public RecordLayer(TLSContentType type) {
		this(type, new ProtocolVersion());
	}
	public RecordLayer(TLSContentType type, ProtocolVersion ver) {
		tls_type = type; 
		opaque_tls_type = tls_type;
		version = ver;
		this.msglist = new ArrayList();
	}
	public RecordLayer(Object msg) {
		this(TLSContentType.HANDSHAKE, msg);
	}

	public RecordLayer(TLSContentType type, Object msg) {
		this(type);
		this.msglist.add(msg);
	}

	public String toString() {
		return "RecordLayer: tls_type " + tls_type + ", opaque_tls_type " + opaque_tls_type + ", ProtocolVersion " + version.major + " " + version.minor + ", length " + length;
	}

	public void print() {
		DebugInfo.printMessage(1,"RecordLayer::print");
		DebugInfo.printMessage(1,"   tls_type " + this.tls_type);
		DebugInfo.printMessage(1,"   opaque_tls_type " + this.opaque_tls_type);
		DebugInfo.printMessage(1,"   length " + this.length);
		if(this.content != null) {
			DebugInfo.printMessage(1,"   content size " + this.content.length);
			for(int i=0; i<this.content.length; i++) {
				if(this.content[i] != null)
					DebugInfo.printMessage(1,"      content " + i + " size "+ this.content[i].length);
			}
		} else
			DebugInfo.printMessage(1,"   content null");
			
		DebugInfo.printMessage(1,"   isLengthSet " + this.isLengthSet);
		DebugInfo.printMessage(1,"   incompleteArrayMessage " + this.incompleteArrayMessage);
		DebugInfo.printMessage(1,"   error " + this.error);
		if(msglist != null)
			DebugInfo.printMessage(1,"   msglist size " + this.msglist.size());
		else
			DebugInfo.printMessage(1,"   msglist null" );
	}

	public int getSize() {
		if(content != null) {
			int contlen = 0;
			for(int i=0; i<content.length; i++){
				if(content[i] != null) 
					contlen += content[i].length;
			}
			return (5 + contlen);
		}
		else
			return 5;
	}
	public void setLength() {
		length = (short)(getSize() - 5);
	}
	public void setLength(int len) {
		length = (short)len;
		isLengthSet = true;
	}
	public short getLength() {
		return length;
	}

	public void headerToByteArray(byte[] ba) {
		ba[0] = opaque_tls_type.type;
		ba[1] = version.major;
		ba[2] = version.minor;
		ba[3] = (byte)(length >>> 8);
		ba[4] = (byte)length;
	}

	public byte[] toByteArrayOnlyHeader() {
		byte[] ba = new byte[5];
		headerToByteArray(ba);
		return ba;
	}

	public byte[] toByteArray() {
		if(isLengthSet == false) setLength();
		byte[] ba = new byte[getSize()];
		headerToByteArray(ba);
		if(content != null) {
			for(int i=0,curpos=5; i<content.length; i++){
				if(content[i] != null) {
					System.arraycopy(content[i], 0, ba, curpos, content[i].length);
					curpos += content[i].length;
				}
			}
		}
		return ba;
	}

	public void setNewContent(byte[][] data) {
		content = data;
	}

	public void setNewContent(byte[] data) {
		content = new byte[1][];
		content[0] = data;
	}

	public void initContent(int len) {
		content = new byte[len][];
		for(int i=0; i<content.length; i++) 
			content[i] = null;
	}

	public void setPlainContent(byte[] text) throws Exception {
		byte[] text_stream = null;
		int len, content_ind, curlen, offset;

		DebugInfo.printMessage(4,"RecordLayer::setPlainContent");
		
		if(text == null) return;
		
		DebugInfo.printDump(4,text,"RecordLayer::setPlainContent input data");

		/*
		 *  create buffer for Content
		 *  
		 *  the length (in bytes) of the TLSPlaintext.fragment MUST NOT exceed 2^14 (=16384)
		 *  the length of the TLSCiphertext.fragment MUST NOT exceed 2^14 + 2^11(=2048)
		 *  
		 *  Number of fragment blocks equals len = (text.length / 16384 + 1)
		 *  TLSCiphertext may add len*2048 bytes, i.e. in blocks (len*2048)/16384 + 1 = len/8 + 1
		 *  All blocks: len + len/8 + 1, where len = (text.length / 16384 + 1)
		 *  
		 *  MAX_LENGTH_TLSInnerPlaintext = 16383 and one byte for 'RecordLayer ContentType'
		 */
		len = (text.length / MAX_LENGTH_TLSInnerPlaintext + 1);
		len += len/8 + 1;
		initContent(len);
		
		if(text.length == 0) {
			content[0] = new byte[0];
			return;
		}
		
		if(text.length >= MAX_LENGTH_TLSInnerPlaintext)
			text_stream = new byte[MAX_LENGTH_TLSInnerPlaintext]; 
		
		content_ind=0; offset=0; 
		len = MAX_LENGTH_TLSInnerPlaintext;
		curlen = text.length;

		do {
			// fill buffer text_stream from text
			if(curlen < MAX_LENGTH_TLSInnerPlaintext) {
				text_stream = new byte[curlen];
				len = curlen;
			}
			System.arraycopy(text, offset, text_stream, 0, len);
			offset += len;
			content[content_ind++] = text_stream;
			curlen = text.length - offset;
		} while(curlen > 0);
	}

	public void setContent(byte[] text, byte[] padding, TLSConnection con) throws Exception {
		byte[] text_stream = null, enc_stream;
		int len, content_ind, curlen, offset;

		DebugInfo.printMessage(4,"RecordLayer::setContent");
		
		if((text == null) || (con == null)) return;
		
		if(!con.isUseCrypto()) {
			this.opaque_tls_type = this.tls_type;
			setPlainContent(text);
			return;
		}
		this.opaque_tls_type = TLSContentType.APPLICATION;
		
		DebugInfo.printDump(4,text,"RecordLayer::setContent input data");

		/*
		 *  create buffer for Content
		 *  
		 *  the length (in bytes) of the TLSPlaintext.fragment MUST NOT exceed 2^14 (=16384)
		 *  the length of the TLSCiphertext.fragment MUST NOT exceed 2^14 + 2^11(=2048)
		 *  
		 *  Number of fragment blocks equals len = (text.length / 16384 + 1)
		 *  TLSCiphertext may add len*2048 bytes, i.e. in blocks (len*2048)/16384 + 1 = len/8 + 1
		 *  All blocks: len + len/8 + 1, where len = (text.length / 16384 + 1)
		 *  
		 *  MAX_LENGTH_TLSInnerPlaintext = 16383 and one byte for 'RecordLayer ContentType'
		 */
		len = (text.length / MAX_LENGTH_TLSInnerPlaintext + 1);
		len += len/8 + 1;
		initContent(len);
		
		if(text.length >= MAX_LENGTH_TLSInnerPlaintext)
			text_stream = new byte[MAX_LENGTH_TLSInnerPlaintext + 1]; // one octet for ContentType  
		
		content_ind=0; offset=0; 
		curlen = text.length;

		do {
			// fill buffer text_stream from text
			if(curlen < MAX_LENGTH_TLSInnerPlaintext) {
				// set padding
				if(padding == null) {
					int blockLength = con.getSecurityParams().block_length;
					if(blockLength != 0)
						padding = con.getSecurityParams().createPadding(blockLength - (text.length + 1) % blockLength);
					else
						padding = new byte[0];
				}

				text_stream = new byte[curlen + 1 + padding.length];
				System.arraycopy(text, offset, text_stream, 0, curlen);
				text_stream[curlen] = this.tls_type.type;
				System.arraycopy(padding, 0, text_stream, curlen + 1, padding.length);
				offset = text.length;
				DebugInfo.printDump(4,text_stream,"RecordLayer::setContent [data,tls type,padding]:");
			}
			else {
				System.arraycopy(text, offset, text_stream, 0, MAX_LENGTH_TLSInnerPlaintext);
				text_stream[MAX_LENGTH_TLSInnerPlaintext] = this.tls_type.type;
				offset += MAX_LENGTH_TLSInnerPlaintext;
			}

			
		// encrypt compressed stream with MAC
			enc_stream = con.encrypt(text_stream);
       		DebugInfo.printDump(5,enc_stream,"RecordLayer::setContent ENCRYPT");
			
			if(enc_stream != null)
				content[content_ind++] = enc_stream;
			
			curlen = text.length - offset;
		} while(curlen > 0);
	}

	public void setContent(byte[][] text, byte[] padding, TLSConnection con) throws Exception {
		setContent(ConvertTypes.combineArrays(text), padding, con);
	}

	// same as setContent, but length may exceed 2^14 (=16384) 
	public void setLargeContent(byte[] text, byte[] padding, TLSConnection con) throws Exception {
		byte[] enc_stream;

		DebugInfo.printMessage(4,"RecordLayer::setLargeContent");
		
		if((text == null) || (con == null)) return;

		if(con.isUseCrypto()) {
			this.opaque_tls_type = TLSContentType.APPLICATION;
			// set padding
			if(padding == null) {
				int blockLength = con.getSecurityParams().block_length;
				if(blockLength != 0)
					padding = con.getSecurityParams().createPadding(blockLength - (text.length + 1) % blockLength);
				else
					padding = new byte[0];
			}
		
			byte[][] plainInnerData = {text, {tls_type.type}, padding};
			DebugInfo.printDump(4,plainInnerData,"RecordLayer::setLargeContent plainInnerData[data,tls type,padding]");
			// encrypt compressed stream with MAC
			enc_stream = con.encrypt(ConvertTypes.combineArrays(plainInnerData));
		} else {
			this.opaque_tls_type = this.tls_type;
			enc_stream = text;
		}

		/*
		 *  The only content block any length
		 */
		initContent(1);
        DebugInfo.printDump(5,enc_stream,"RecordLayer::setLargeContent ENCRYPT");
		if(enc_stream != null)
			content[0] = enc_stream;
	}

	// same as setContent, but length may exceed 2^14 (=16384) 
	public void setLargeContent(byte[][] text, byte[] padding, TLSConnection con) throws Exception {
		setLargeContent(ConvertTypes.combineArrays(text), padding, con);
	}

	public void setContent(Object msg, byte[] padding, TLSConnection con) throws Exception {
		setContent(prepairContent(msg), padding, con);
	}

	public void setLargeContent(Object msg, byte[] padding, TLSConnection con) throws Exception {
		setLargeContent(prepairContent(msg), padding, con);
	}

	public void setContent(Object msg, byte[] padding, TLSConnection con, TLSContentType type) throws Exception {
		byte[] data = prepairContent(msg);
		tls_type = type;
		setContent(data, padding, con);
	}

	public byte[] prepairContent(Object msg) throws Exception {
		byte[] plainData = null;
		if(msg != null) {
			if(msg instanceof HandshakeMessage) {
				tls_type = TLSContentType.HANDSHAKE;
				if( ((HandshakeMessage)msg).net_msg == null)
					((HandshakeMessage)msg).toByteArray();
				plainData = ((HandshakeMessage)msg).net_msg;
			}
			else if(msg instanceof Alert) {
				tls_type = TLSContentType.ALERT;
				plainData = ((Alert)msg).toByteArray();
			}
			else if(msg instanceof ChangeCipher) {
				tls_type = TLSContentType.CHANGE_CIPHER_SPEC;
				plainData = ((ChangeCipher)msg).toByteArray();
			}
			else if(msg instanceof ApplicationData) {
				tls_type = TLSContentType.APPLICATION;
				plainData = ((ApplicationData)msg).data;
			}
			else if(msg instanceof byte[]) {
				plainData = (byte[])msg;
			}
			else if(msg instanceof byte[][]) {
				plainData = ConvertTypes.combineArrays((byte[][])msg);
			}
		}
		return plainData;
	}

	public void setContent(Class cl, ArrayList msglist, int offset, int len, byte[] padding, TLSConnection con) throws Exception {
		int data_ind, i, curlen;
		
		if((cl != null)&&(msglist != null)&&(offset < msglist.size())) {
			Object msg;
			for(data_ind=0,i=offset,curlen=len;i < msglist.size() && curlen>0; i++,curlen--) {
				msg = msglist.get(i);
				if(cl.isAssignableFrom(msg.getClass())) data_ind++;
			}
			if(data_ind == 0) return;
			
			DebugInfo.printMessage(3,"RecordLayer::setContent(Class): find " + data_ind + " messages");
			byte[][] data = new byte[data_ind][];
			i=offset; data_ind=0;
			
			if(HandshakeMessage.class.isAssignableFrom(cl)) {
				tls_type = TLSContentType.HANDSHAKE;
				for(;i < msglist.size() && len>0; i++,len--) {
					msg = msglist.get(i);
					if(msg instanceof HandshakeMessage) {
						if( ((HandshakeMessage)msg).net_msg == null)
							((HandshakeMessage)msg).toByteArray();
						data[data_ind++] = ((HandshakeMessage)msg).net_msg;
					}
				}
			}
			else if(Alert.class.isAssignableFrom(cl)) {
				tls_type = TLSContentType.ALERT;
				for(;i < msglist.size() && len>0; i++,len--) {
					msg = msglist.get(i);
					if(msg instanceof Alert) {
						data[data_ind++] = ((Alert)msg).toByteArray();
					}
				}
			}
			else if(ChangeCipher.class.isAssignableFrom(cl)) {
				tls_type = TLSContentType.CHANGE_CIPHER_SPEC;
				for(;i < msglist.size() && len>0; i++,len--) {
					msg = msglist.get(i);
					if(msg instanceof ChangeCipher) {
						data[data_ind++] = ((ChangeCipher)msg).toByteArray();
					}
				}
			}
			else if(ApplicationData.class.isAssignableFrom(cl)) {
				tls_type = TLSContentType.APPLICATION;
				for(;i < msglist.size() && len>0; i++,len--) {
					msg = msglist.get(i);
					if(msg instanceof ApplicationData) {
						data[data_ind++] = ((ApplicationData)msg).data;
					}
				}
			}
			else
				return;
			
			setContent(data, padding, con);
		}
	}

	public void printMessages(int debugLevel) {
		DebugInfo.printMessage(debugLevel, toString());
		Object msg;
		if(msglist != null  &&  !msglist.isEmpty()) {
			for(int i=0; i < msglist.size(); i++) {
				msg = msglist.get(i);
				if( (msg instanceof HandshakeMessage)
				 || (msg instanceof Alert)
				 || (msg instanceof ChangeCipher)
				 || (msg instanceof ApplicationData)) {
						DebugInfo.printMessage(debugLevel, "   " + msg.toString());
				}
				else
					DebugInfo.printMessage(debugLevel, "   message " + i + ": unknown type");
			}
		}
	}

    // decrypt one TLS message
    public int getTlsPlain(TLSConnection con) throws Exception {
    	error = 0;
    	if(content != null && content[0] != null) {
 			content[0] = con.decrypt(content[0]);
   			if(content[0] == null) {
   				error = -1;
   				DebugInfo.printMessage(2,"===Error! RecordLayer::getTlsPlain: Decrypt message failed");
   			}
    	}
    	else {
				error = -2;
   				DebugInfo.printMessage(3,"===Error! RecordLayer::getTlsPlain: Bad input data");
    	}
    	return error;
    }
    
    //  check and trim Padding
    public int trimPadding() throws Exception {
    	error = 0; 
    	byte[] data;
    	int len;
    	
    	if(content != null && content[0] != null) {
    		data = content[0];
    		DebugInfo.printDump(4,data,null);
   				len = data.length - 1;
   				while(len >= 0) {
   				// check Padding
   					if(data[len] != 0) {
   						this.opaque_tls_type = this.tls_type;
   						this.tls_type = TLSContentType.valueOf(data[len]);
   						this.length = (short)len;
   						content[0] = new byte[len];
 						System.arraycopy(data, 0, content[0], 0, len);
 						break;
   					}
   					len--;
   				}

   				// if no data
   				if(len < 0) {
   	                content[0] = new byte[0];
   					error = -4;
   					DebugInfo.printMessage(3,"===Error! RecordLayer::trimPadding: no plain data, only padding");
   				}
    	}
    	else {
			error = -5;
			DebugInfo.printMessage(3,"===Error! RecordLayer::trimPadding: Bad input data");
    	}
    	return error;
    }
    
    // parse one TLS message
    public int parseTlsPlain(TLSConnection con) {
    	error = 0;
    	int j;
   		msglist = new ArrayList();

   		if(content != null && content[0] != null) {
    		DebugInfo.printDump(4,content[0],null);
   			switch(tls_type) {
   				case CHANGE_CIPHER_SPEC:
   					if(content[0].length != 1) {
   						DebugInfo.printDump(1,content[0],"===Error! RecordLayer::parseTlsPlain: CHANGE_CIPHER_SPEC long data");
   						error = -1;
   					}
   					for(j=0; j < content[0].length; j++)
   						msglist.add(new ChangeCipher(content[0][j]));
   					break;
   				case ALERT:
   					int len = content[0].length;
   					if(len % 2 != 0) {
   						len--;
   						DebugInfo.printMessage(1,"===Error! RecordLayer::parseTlsPlain: ALERT incomplete data");
   						error = -1;
   					}
   					for(j=0; j < len; )
   						msglist.add(new Alert(content[0][j++],content[0][j++]));
   					break;
   				case HANDSHAKE:
   					error = HandshakeMessage.parseHandshakeMessagesList(content[0], msglist, con);
   					break;
   				case APPLICATION:
					msglist.add(new ApplicationData(content[0]));
   					break;
   				default:	
					DebugInfo.printMessage(3,"===Error! RecordLayer::parseTlsPlain: unknown message type " + tls_type);
   					error = -1;
   			}
   		}
   		return error;
    }

	public int decryptAndParseTLS(TLSConnection con) {
		int res = -1;
		DebugInfo.printMessage(4,">>> RecordLayer::decryptAndParseTLS");

		if(con == null) {
			DebugInfo.printMessage(4,"connection not specified");
			return res;
		}
		
		if(version.major != 3) {
			DebugInfo.printMessage(1,"===Error! Bad major TLS version: " + version.major + " " + version.minor);
			DebugInfo.printDump(1,toByteArrayOnlyHeader(),"TLS header:");
			DebugInfo.printDump(1,content[0],"TLS data:");
    		res = 70;
   			return res;
		}

		if(version.minor != 3) {
			DebugInfo.printMessage(1,"===Warning! Bad minor TLS version: " + version.major + " " + version.minor);
			DebugInfo.printDump(1,toByteArrayOnlyHeader(),"TLS header:");
			DebugInfo.printDump(1,content[0],"TLS data:");
			if((tls_type != TLSContentType.ALERT) && (tls_type != TLSContentType.HANDSHAKE)) {
				res = 70;
				return res;
			}
		}

		if(con.isUseCrypto() && 
		   tls_type != TLSContentType.CHANGE_CIPHER_SPEC && 
		   tls_type != TLSContentType.ALERT) {
			if(tls_type != TLSContentType.APPLICATION)
				DebugInfo.printMessage(1,"===Warning! For crypto messages TLSContentType must be APPLICATION: " + tls_type);

			// decrypt TLS record
			DebugInfo.printMessage(4,"try decrypt message");
			try {
				getTlsPlain(con);
				if(error != 0) {
					return 51;
				}
			}
			catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! RecordLayer::getTlsPlain: " + err);
				err.printStackTrace();
				return 51;
			}

			// check and trim Padding
			DebugInfo.printMessage(3,"trim padding");
			try {
				trimPadding();
				if(error != 0) {
	    			return 20;
				}
			}
			catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! RecordLayer::trimPaddingAndMAC: " + err);
				return 20;
			}
		}
		
    	// parse TLS record
		DebugInfo.printMessage(4,"parse message");
    	parseTlsPlain(con);
    	if(error != 0) {
    		res = 50;
    		return res;
    	}

    	printMessages(2);
    	return res;
	}


    public byte[] getApplicationData() throws Exception {
     	int j,len = 0,pos = 0;
     	byte[] data = null;
     	ApplicationData curAppDt;
   		
   		if(this.tls_type == TLSContentType.APPLICATION) {
   			for(j=0; j< this.msglist.size(); j++) {
   				curAppDt = (ApplicationData)(this.msglist.get(j));
   				if(curAppDt.data != null)
   					len += curAppDt.data.length;
   			}
   			
   			data = new byte[len];
   			
   			for(j=0; j< this.msglist.size(); j++) {
   				curAppDt = (ApplicationData)(this.msglist.get(j));
   				if(curAppDt.data != null) {
   	   				System.arraycopy(curAppDt.data, 0, data, pos, curAppDt.data.length);
   	   				pos += curAppDt.data.length;
   				}
   			}
   		}
   		
   		return data;
    }

    public static byte[] getApplicationData(ArrayList<RecordLayer> tlsResp) throws Exception {
     	int j,len = 0,pos = 0;
     	byte[] data = null, curDt;
     	ArrayList<byte[]> msglist = new ArrayList<byte[]>();
   		
   		if(tlsResp != null) {
   			for(j=0; j< tlsResp.size(); j++) {
   				msglist.add(tlsResp.get(j).getApplicationData());
   			}

   			for(j=0; j< msglist.size(); j++) {
   				curDt = msglist.get(j);
   				if(curDt != null)
   					len += curDt.length;
   			}
   			
   			data = new byte[len];
   			
   			for(j=0; j< msglist.size(); j++) {
   				curDt = msglist.get(j);
   				if(curDt != null) {
   	   				System.arraycopy(curDt, 0, data, pos, curDt.length);
   	   				pos += curDt.length;
   				}
   			}
   		}

   		return data;
    }
}
