package ru.ispras.telecomtesk.tlsv13.TLSAlgorithms;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;

public enum SignatureScheme {
    /* RSASSA-PKCS1-v1_5 algorithms */
    RSA_PKCS1_SHA256(0X0401),
    RSA_PKCS1_SHA384(0X0501),
    RSA_PKCS1_SHA512(0X0601),

    /* ECDSA algorithms */
    ECDSA_SECP256R1_SHA256(0X0403),
    ECDSA_SECP384R1_SHA384(0X0503),
    ECDSA_SECP521R1_SHA512(0X0603),

    /* RSASSA-PSS algorithms with public key OID rsaEncryption */
    RSA_PSS_RSAE_SHA256(0X0804),
    RSA_PSS_RSAE_SHA384(0X0805),
    RSA_PSS_RSAE_SHA512(0X0806),

    /* EdDSA algorithms */
    ED25519(0X0807),
    ED448(0X0808),

    /* RSASSA-PSS algorithms with public key OID RSASSA-PSS */
    RSA_PSS_PSS_SHA256(0X0809),
    RSA_PSS_PSS_SHA384(0X080A),
    RSA_PSS_PSS_SHA512(0X080B),

    /* Legacy algorithms */
    RSA_PKCS1_SHA1(0X0201),
    ECDSA_SHA1(0X0203),

    /* Reserved Code Points 0XFE00..0XFFFF */
    PRIVATE_USE(0XFE00),
	UNKNOWN(0xFFFF);
	
	public int scheme;
	
	private SignatureScheme(int alg) {
		scheme = alg;
	}
	public int getValue(){
		return scheme;
	}
	public static SignatureScheme valueOf(int scheme) {
		switch (scheme) {
		case 0X0401: return RSA_PKCS1_SHA256;
		case 0X0501: return RSA_PKCS1_SHA384;
		case 0X0601: return RSA_PKCS1_SHA512;
		case 0X0403: return ECDSA_SECP256R1_SHA256;
		case 0X0503: return ECDSA_SECP384R1_SHA384;
		case 0X0603: return ECDSA_SECP521R1_SHA512;
		case 0X0804: return RSA_PSS_RSAE_SHA256;
		case 0X0805: return RSA_PSS_RSAE_SHA384;
		case 0X0806: return RSA_PSS_RSAE_SHA512;
		case 0X0807: return ED25519;
		case 0X0808: return ED448;
		case 0X0809: return RSA_PSS_PSS_SHA256;
		case 0X080A: return RSA_PSS_PSS_SHA384;
		case 0X080B: return RSA_PSS_PSS_SHA512;
		case 0X0201: return RSA_PKCS1_SHA1;
		case 0X0203: return ECDSA_SHA1;
		default: return UNKNOWN;
		}
	}

	public String getHashName() {
		switch (this) {
		case RSA_PKCS1_SHA256: 
		case ECDSA_SECP256R1_SHA256: 
		case RSA_PSS_RSAE_SHA256: 
		case RSA_PSS_PSS_SHA256: 
			return "SHA256";
		case RSA_PKCS1_SHA384: 
		case ECDSA_SECP384R1_SHA384: 
		case RSA_PSS_RSAE_SHA384: 
		case RSA_PSS_PSS_SHA384: 
			return "SHA384";
		case RSA_PKCS1_SHA512: 
		case ECDSA_SECP521R1_SHA512: 
		case RSA_PSS_RSAE_SHA512: 
		case RSA_PSS_PSS_SHA512: 
			return "SHA512";
		case RSA_PKCS1_SHA1: 
		case ECDSA_SHA1: 
			return "SHA1";
		case ED25519: 
		case ED448: 
		default: return "UNKNOWN";
		}
	}

	public String getSigName() {
		switch (this) {
		case ECDSA_SECP256R1_SHA256: 
		case ECDSA_SECP384R1_SHA384: 
		case ECDSA_SECP521R1_SHA512: 
		case ECDSA_SHA1: 
			return "ECDSA";
		case RSA_PKCS1_SHA256: 
		case RSA_PSS_RSAE_SHA256: 
		case RSA_PSS_PSS_SHA256: 
		case RSA_PKCS1_SHA384: 
		case RSA_PSS_RSAE_SHA384: 
		case RSA_PSS_PSS_SHA384: 
		case RSA_PKCS1_SHA512: 
		case RSA_PSS_RSAE_SHA512: 
		case RSA_PSS_PSS_SHA512: 
		case RSA_PKCS1_SHA1: 
			return "RSA";
		case ED25519: 
		case ED448: 
			return "ED";
		default: return "UNKNOWN";
		}
	}

	public String toString() {
		return ConvertTypes.codeToString(this.scheme, this.name());
	}

	public static String toString(int scheme) {
		return ConvertTypes.codeToString(scheme, SignatureScheme.valueOf(scheme).name());
	}
}
