package ru.ispras.telecomtesk.tlsv13.TLSState;

import java.net.Socket;
import java.util.ArrayList;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.*;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.*;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.HandshakeMessageType;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.KeyExchangeMode;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.NamedGroup;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.TLSCipherSuite;

public class TLSConnection 
{
	public Socket skt;
	public TLSConnectionSelector selector;
	public TLSHandshakeParameters hskParams;
	public TLSSecurityParameters tlsParams;
	public ConnectionRole localRole;
	public enum StateDirection {READ_STATE, WRITE_STATE};

	public TLSConnection(TLSConnectionSelector sel) {
		selector = sel;
		localRole = ConnectionRole.CLIENT;
		hskParams = new TLSHandshakeParameters();
		tlsParams = new TLSSecurityParameters();
	}
	public TLSConnection(TLSConnectionSelector sel, ConnectionRole isClient) {
		this(sel);
		this.localRole = isClient;
	}
	public void init() {
		hskParams = new TLSHandshakeParameters();
		tlsParams = new TLSSecurityParameters();
	}
	public void setSocketParameters(Socket skt) {
		if(skt != null) {
			this.skt = skt;
			selector = new TLSConnectionSelector(
					skt.getLocalAddress(),
					skt.getInetAddress(),
					skt.getLocalPort(),
					skt.getPort()
					);
		}
	}
	public void setAAD(byte [] aad) {
		if(tlsParams != null) 
			tlsParams.presetAAD = aad;
	}
	public TLSSecurityParameters getSecurityParams() {
		return tlsParams;
	}
	public TLSHandshakeParameters getHandshakeParams() {
		return hskParams;
	}

	public void incrementSenderSequenceNumber() {
		if(localRole == ConnectionRole.CLIENT)
			tlsParams.clientKey.incrementSequenceNumber();
		else
			tlsParams.serverKey.incrementSequenceNumber();
	}
	public void incrementReceiverSequenceNumber() {
		if(localRole == ConnectionRole.SERVER)
			tlsParams.clientKey.incrementSequenceNumber();
		else
			tlsParams.serverKey.incrementSequenceNumber();
	}

	public void decrementSenderSequenceNumber() {
		if(localRole == ConnectionRole.CLIENT)
			tlsParams.clientKey.decrementSequenceNumber();
		else
			tlsParams.serverKey.decrementSequenceNumber();
	}
	public void decrementReceiverSequenceNumber() {
		if(localRole == ConnectionRole.SERVER)
			tlsParams.clientKey.decrementSequenceNumber();
		else
			tlsParams.serverKey.decrementSequenceNumber();
	}

	public byte[] getHandshakeTrafficSecret(MessageDirection msgDir) {
		switch(msgDir) {
		case CLIENT_SEND:
		case SERVER_RECEIVE:
			return tlsParams.client_handshake_traffic_secret;
		case SERVER_SEND: 
		case CLIENT_RECEIVE:
			return tlsParams.server_handshake_traffic_secret;
		}
		return null;
	}

	public void printDump(int debugLevel) {
		DebugInfo.printMessage(2,"TLSConnection::print: Role " + localRole);
		tlsParams.printDump(debugLevel);
		hskParams.printDump(debugLevel);
	}
	
	public boolean isUseCrypto() {
		return tlsParams.isUseCrypto(); 
	}

	public byte[] encrypt(byte[] data) throws Exception {
		return tlsParams.encrypt(data, localRole);
	}
	
	public byte[] encrypt(byte[][] data) throws Exception {
		return tlsParams.encrypt(data, localRole);
	}
	
	public byte[] decrypt(byte[] data) throws Exception {
		return tlsParams.decrypt(data, localRole);
	}
	
	public int updateState(HelloRetryRequest msg, Hello last_client_hello) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: HelloRetryRequest");
		hskParams.isTLS13 = false;
			try {
				if(msg.extensions != null) {
					for(TLSExtension ext : msg.extensions) {
						switch(ext.getExtensionType()) {
							case SUPPORTED_VERSIONS:
								hskParams.isTLS13 = ((SupportedVersionsExt)ext).isTLS13();
								break; 
							case KEY_SHARE: 
								hskParams.serverSupportedGroups = new NamedGroup[1];
								hskParams.serverSupportedGroups[0] = ((KeyShareHelloRetry)ext).data;
								break; 
							case COOKIE: 
								hskParams.remoteCookie = ((CookieExt)ext).data;
								break; 
							default: ; 
						}
					}
				}

			}
			catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! TLSConnection::updateState:HelloRetryRequest: " + err);
				err.printStackTrace();
			}

			try {
				if(last_client_hello != null) {
					byte[] chNetMsg = last_client_hello.net_msg;
					byte[] chHash = TLSKeyDerivation.hash(TLSCipherSuite.valueOf(msg.cipher_suites[0]).getMacAlgorithm(), chNetMsg);
					chNetMsg = new byte[chHash.length + 4];
					chNetMsg[0] = HandshakeMessageType.MESSAGE_HASH.getValue();
					chNetMsg[1] = 0;
					chNetMsg[2] = 0;
					chNetMsg[3] = (byte)chHash.length;
					System.arraycopy(chHash, 0, chNetMsg, 4, chHash.length);
					last_client_hello.net_msg = chNetMsg;
				}
			} catch (Exception err) {}
			
		return 0;
	}

	public int updateState(Hello msg) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: Hello");
		int res = 0;
		hskParams.isTLS13 = false;
		if(msg.msg_type == 1) {
			hskParams.clientCipherSuites = msg.cipher_suites;
			tlsParams.client_random = msg.rnd;
			try {
				DebugInfo.printMessage(5,"TLSConnection::updateState:Hello extensions");
				if(msg.extensions != null) {
					for(TLSExtension ext : msg.extensions) {
						switch(ext.getExtensionType()) {
							case SUPPORTED_VERSIONS:
								hskParams.isTLS13 = ((SupportedVersionsExt)ext).isTLS13();
								break; 
							case SUPPORTED_GROUPS: 
								hskParams.clientSupportedGroups = ((SupportedGroupsExt)ext).data;
								break; 
							case SIGNATURE_ALGORITHMS: 
								hskParams.clientSignatureScheme = ((SignatureAndHashAlgorithmExt)ext).data;
								break; 
							case KEY_SHARE: 
								hskParams.clientKeyShareList = ((KeyShareClientHello)ext).data;
								break; 
							case POST_HANDSHAKE_AUTH: 
								hskParams.isClientPostAuth = true;
								break; 
							default: ; 
						}
					}
				}
			}
			catch (Exception ignored) {res = 1;}
		} else {
			tlsParams.setAlgorithmParams(TLSCipherSuite.valueOf(msg.cipher_suites[0]));
			tlsParams.server_random = msg.rnd;
			hskParams.keyExchangeMode = KeyExchangeMode.DHE;
			try {
				if(msg.extensions != null) {
					for(TLSExtension ext : msg.extensions) {
						switch(ext.getExtensionType()) {
							case SUPPORTED_VERSIONS:
								hskParams.isTLS13 = ((SupportedVersionsExt)ext).isTLS13();
								break; 
							case SUPPORTED_GROUPS: 
								hskParams.serverSupportedGroups = ((SupportedGroupsExt)ext).data;
								break; 
							case SIGNATURE_ALGORITHMS: 
								hskParams.serverSignatureScheme = ((SignatureAndHashAlgorithmExt)ext).data;
								break; 
							case KEY_SHARE: 
								hskParams.serverKeyShare = ((KeyShareServerHello)ext).data;
								break; 
							case PSK_KEY_EXCHANGE_MODES:
								hskParams.keyExchangeMode = KeyExchangeMode.PSK;
								break; 
							case EARLY_DATA:
								hskParams.mustSendEndOfEarlyData = true;
								break; 
							default: ; 
						}
					}
				}

				if (tlsParams.early_secret == null) {
					tlsParams.early_secret = TLSKeyDerivation.TLS_HKDF_Extract(tlsParams.mac_alg, null, tlsParams.psk);
				}
				tlsParams.dheSecret = hskParams.createDHEsecret(localRole);
				byte[] tmpDerive = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.early_secret, "derived");
				tlsParams.handshake_secret = TLSKeyDerivation.TLS_HKDF_Extract(tlsParams.mac_alg, tmpDerive, tlsParams.dheSecret);
				hskParams.sessionMsgList.add(msg);
				byte[] messagesHash = TLSKeyDerivation.TLS_TranscriptHash(tlsParams.mac_alg, hskParams.sessionMsgList);
				tlsParams.client_handshake_traffic_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.handshake_secret, "c hs traffic", messagesHash);
				tlsParams.server_handshake_traffic_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.handshake_secret, "s hs traffic", messagesHash);
				tlsParams.createHandshakeTrafficKeys();
				hskParams.sessionMsgList.remove(msg);
				DebugInfo.printMessage(5,"TLSConnection::updateState:Hello tls keys created");
				tlsParams.printDump(5);
			}
			catch (Exception err) {
				DebugInfo.printMessage(1,"===Error! TLSConnection::updateState:Hello: " + err);
				err.printStackTrace();
			}
		}
		return res;
	}

	public int updateState(EncryptedExtensions msg) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: EncryptedExtensions");
		for(TLSExtension ext : msg.getExtensions()) {
			switch(ext.getExtensionType()) {
				case SUPPORTED_GROUPS: 
					hskParams.serverSupportedGroups = ((SupportedGroupsExt)ext).data;
					break; 
				default: ; 
			}
		}
		return 0;
	}

	public int updateState(CertificateRequest msg) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: CertificateRequest");
		hskParams.certReq = msg;
		return 0;
	}

	public int updateState(HMCertificate msg) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: HMCertificate");
		return 0;
	}

	public int updateState(Finished msg, MessageDirection msgDir) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: Finished");
		try {
    		switch(msgDir) {
    		case CLIENT_SEND:
    		case SERVER_RECEIVE:
    			hskParams.sessionMsgList.add(msg);
    			tlsParams.resumption_master_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.master_secret, "res master", hskParams.sessionMsgList);
    			hskParams.sessionMsgList.remove(msg);
    			tlsParams.createClientApplicationTrafficKeys();
    			break;
    		case SERVER_SEND: 
    		case CLIENT_RECEIVE:
    			byte[] tmpDerive = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.handshake_secret, "derived");
    			tlsParams.master_secret = TLSKeyDerivation.TLS_HKDF_Extract(tlsParams.mac_alg, tmpDerive, null);
    			hskParams.sessionMsgList.add(msg);
				byte[] messagesHash = TLSKeyDerivation.TLS_TranscriptHash(tlsParams.mac_alg, hskParams.sessionMsgList);
    			tlsParams.client_application_traffic_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.master_secret, "c ap traffic", messagesHash);
    			tlsParams.server_application_traffic_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.master_secret, "s ap traffic", messagesHash);
    			tlsParams.exporter_master_secret = TLSKeyDerivation.TLS_DeriveSecret(tlsParams.mac_alg, tlsParams.master_secret, "exp master", hskParams.sessionMsgList);
    			hskParams.sessionMsgList.remove(msg);
    			tlsParams.createServerApplicationTrafficKeys();
            	break;
    		}
			DebugInfo.printMessage(5,"TLSConnection::updateState: Finished tls keys created");
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! TLSConnection::updateState: Finished: " + err);
			err.printStackTrace();
		}

		tlsParams.printDump(5);
		return 0;
	}

	public int updateState(KeyUpdate msg, MessageDirection msgDir) {
		DebugInfo.printMessage(4,"TLSConnection::updateState: KeyUpdate");
		try {
    		switch(msgDir) {
    		case CLIENT_SEND:
    		case SERVER_RECEIVE:
    			tlsParams.updateClientTrafficKeys();
    			break;
    		case SERVER_SEND: 
    		case CLIENT_RECEIVE:
    			tlsParams.updateServerTrafficKeys();
            	break;
    		}
			DebugInfo.printMessage(5,"TLSConnection::updateState: KeyUpdate tls keys created");
		} catch (Exception err) {
			DebugInfo.printMessage(1,"===Error! TLSConnection::updateState: KeyUpdate: " + err);
			err.printStackTrace();
		}
		return 0;
	}

	public void setRandom(byte[] rnd, boolean isClientRandom) {
		tlsParams.setRandom(rnd, isClientRandom);
	}
	public void setAlgorithmParams(TLSCipherSuite cs) {
		tlsParams.setAlgorithmParams(cs);
	}
	public void createMasterSecret(byte[] pre_master_secret) throws Exception {
		tlsParams.createMasterSecret(pre_master_secret);
		tlsParams.setMasterSecret(tlsParams.getMasterSecret());
	}
	public void setMasterSecret(byte[] ms) {
		tlsParams.setMasterSecret(ms);
	}

	public int createSecretKeys() throws Exception {
		int len,res,pos=0;
		byte[] buf;
		byte[] rnd = new byte[64];
		TLSStateKeys cl, srv;
		
		len = (tlsParams.enc_key_length + tlsParams.mac_key_length + tlsParams.fixed_iv_length)*2;
		buf = new byte[len];
		System.arraycopy(tlsParams.server_random, 0, rnd, 0, 32);
		System.arraycopy(tlsParams.client_random, 0, rnd, 32, 32);
		res = tlsParams.prf(tlsParams.master_secret, "key expansion", rnd, buf);

		cl = new TLSStateKeys(tlsParams.mac_key_length, tlsParams.enc_key_length, tlsParams.fixed_iv_length);
		srv = new TLSStateKeys(tlsParams.mac_key_length, tlsParams.enc_key_length, tlsParams.fixed_iv_length);

		if(tlsParams.mac_key_length != 0) {
			System.arraycopy(buf, pos, cl.macKey, 0, tlsParams.mac_key_length);
			pos += tlsParams.mac_key_length;
			System.arraycopy(buf, pos, srv.macKey, 0, tlsParams.mac_key_length);
			pos += tlsParams.mac_key_length;
		}
		if(tlsParams.enc_key_length != 0) {
			System.arraycopy(buf, pos, cl.encKey, 0, tlsParams.enc_key_length);
			pos += tlsParams.enc_key_length;
			System.arraycopy(buf, pos, srv.encKey, 0, tlsParams.enc_key_length);
			pos += tlsParams.enc_key_length;
		}
		if(tlsParams.fixed_iv_length != 0) {
			System.arraycopy(buf, pos, cl.fixedIV, 0, tlsParams.fixed_iv_length);
			pos += tlsParams.fixed_iv_length;
			System.arraycopy(buf, pos, srv.fixedIV, 0, tlsParams.fixed_iv_length);
		}
		
                    
		return res;
	}

}



