package ru.ispras.telecomtesk.tlsv13.TLSState;

import java.math.BigInteger;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.cert.Certificate;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;
import java.util.ArrayList;
import java.util.Arrays;

import javax.crypto.KeyAgreement;
import javax.crypto.interfaces.DHPrivateKey;
import javax.crypto.interfaces.DHPublicKey;
import javax.crypto.spec.DHPublicKeySpec;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.CertificateRequest;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.HandshakeMessage;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.TLSExtensions.KeyShareEntry;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;

public class TLSHandshakeParameters
{
	public boolean isTLS13 = false;
	public boolean mustSendEndOfEarlyData = false;
	public boolean isClientPostAuth = false;
	public short[] clientCipherSuites = null; // 2..2^16-2 bytes
	public KeyExchangeMode keyExchangeMode = KeyExchangeMode.NULL; 
	public NamedGroup[] clientSupportedGroups = null;
	public NamedGroup[] serverSupportedGroups = null;
	public SignatureScheme[] clientSignatureScheme = null;
	public SignatureScheme[] serverSignatureScheme = null;
	public byte[] remoteCookie = null;
	public ArrayList<KeyShareEntry> clientKeyShareList = null;
	public ArrayList<KeyPair> localKeyPairList = null;
	public ArrayList sessionMsgList = new ArrayList();
	public KeyShareEntry clientKeyShare = null;
	public KeyShareEntry serverKeyShare = null;
	public KeyPair localKeyPair = null;
	public X509Certificate remoteCert = null;
	public Certificate localCert = null;
	public SignatureScheme localCertType = SignatureScheme.UNKNOWN;
	public CertificateRequest certReq = null;

	public TLSHandshakeParameters() {
		isTLS13 = false;
	}

	// suppose clientCipherSuites != null
	private boolean checkCipherSuite(short cs) {
		boolean res = false;
		for(short csCur : clientCipherSuites) {
			if(csCur == cs) {
				res = true;
				break;
			}
		}
		return res;
	}
	
	public short[] chooseCipherSuite(short[] serverCipherSuites) {
		short[] res = null;
		if((clientCipherSuites != null)&&(serverCipherSuites != null))
			for(short serCS : serverCipherSuites) {
				if(checkCipherSuite(serCS)) {
					res = new short[]{serCS};
					break;
				}
			}
		return res;
	}
	
	// suppose clientSupportedGroups != null
	private boolean checkNamedGroup (NamedGroup ng) {
		boolean res = false;
		for(NamedGroup ngCur : clientSupportedGroups) {
			if(ngCur == ng) {
				res = true;
				break;
			}
		}
		return res;
	}
	
	public NamedGroup[] chooseNamedGroup(NamedGroup[] serverNG) {
		NamedGroup[] res = null;
		if((clientSupportedGroups != null)&&(serverNG != null))
			for(NamedGroup serNG : serverNG) {
				if(checkNamedGroup(serNG)) {
					res = new NamedGroup[]{serNG};
					break;
				}
			}
		return res;
	}

	// suppose clientKeyShareList != null
	private boolean checkKeyShare (NamedGroup ng) {
		boolean res = false;
		for(KeyShareEntry curKS : clientKeyShareList) {
			if((curKS != null)&&(curKS.getGroup() == ng)) {
				res = true;
				break;
			}
		}
		return res;
	}
	
	public KeyShareEntry getFirstClientKeyShare() {
		if(clientKeyShareList != null)
			return clientKeyShareList.get(0);
		else
			return null;
	}
	
	public NamedGroup[] chooseKeyShare(NamedGroup[] serverNG) {
		NamedGroup[] res = null;
		if((clientKeyShareList != null)&&(serverNG != null))
			for(NamedGroup serNG : serverNG) {
				if(checkKeyShare(serNG)) {
					res = new NamedGroup[]{serNG};
					break;
				}
			}
		return res;
	}
	
	public NamedGroup[] chooseRetryNamedGroup(NamedGroup[] serverNG) {
		NamedGroup[] res = null;
		if((clientKeyShareList != null)&&(clientSupportedGroups != null)&&(serverNG != null))
			for(NamedGroup serNG : serverNG) {
				if(!checkKeyShare(serNG) && checkNamedGroup(serNG)) {
					res = new NamedGroup[]{serNG};
					break;
				}
			}
		return res;
	}
	
	// generate key pairs list on 'CLIENT' side
	public void generateClientKeyPairList(NamedGroup[] hgList) {
		KeyPair kp = null;
		ArrayList<KeyShareEntry> kseList = new ArrayList<KeyShareEntry>();
		ArrayList<KeyPair> kpList = new ArrayList<KeyPair>();

		if(hgList != null) {
			for(int i=0; i<hgList.length; i++) {
				try {
					kp = TLSKeyDerivation.generateKeyPair(hgList[i]);
					PublicKey pubKey = kp.getPublic();
					kseList.add(new KeyShareEntry(hgList[i], pubKey));
                    // if an error will be, then 'kseList' and 'kpList' will not changed
					kpList.add(kp);
				} catch (Exception e) {
					DebugInfo.printMessage(1, "!!! Error: TLSHandshakeParameters::generateClientKeyPairList: failed for group " + hgList[i]);
				}
			}
		}
		this.clientKeyShareList = kseList;
		this.localKeyPairList = kpList;
	}

	// generate key pair on 'SERVER' side
	public void generateServerKeyPair(NamedGroup group) {
		try {
			localKeyPair = TLSKeyDerivation.generateKeyPair(group);

			PublicKey pubKey = localKeyPair.getPublic();
			serverKeyShare = new KeyShareEntry(group, pubKey);
		} catch (Exception e) {
			DebugInfo.printMessage(1, "!!! Error: TLSHandshakeParameters::generateServerKeyPair: failed for group " + group);
		}
	}

	public byte[] createDHEsecret(ConnectionRole direction) {
		DebugInfo.printMessage(4,"TLSHandshakeParameters::createDHEsecret");
		byte[] DHsecret = null;
		KeyShareEntry remoteKeyShare = null;
		BigInteger publicKeyY = null;
		
		try {
			if(direction == ConnectionRole.CLIENT) {
				localKeyPair = localKeyPairList.get(KeyShareEntry.getPairKeyShareIndex(clientKeyShareList, serverKeyShare));
				remoteKeyShare = serverKeyShare;
			} else {
				if(clientKeyShare == null)
					clientKeyShare = KeyShareEntry.getPairKeyShareEntry(clientKeyShareList, serverKeyShare);
				remoteKeyShare = clientKeyShare;
			}

			DebugInfo.printMessage(5,"TLSHandshakeParameters::createDHEsecret: create SharedSecret");
			PrivateKey localDHPrivateKey = localKeyPair.getPrivate();
			PublicKey remoteDHPublicKey = null;
			KeyAgreement keyAgree = null;
			NamedGroup ng = remoteKeyShare.getGroup();
			
			if(ng.isECDHE()) {
			    ECPoint remoteP = remoteKeyShare.getECPoint();
		        ECPublicKey localPublicKey = (ECPublicKey)localKeyPair.getPublic(); 
				ECPublicKeySpec ecKeySpec = new ECPublicKeySpec(remoteP, localPublicKey.getParams());
				KeyFactory localKeyFactory = KeyFactory.getInstance("EC");
				remoteDHPublicKey = (ECPublicKey) localKeyFactory.generatePublic(ecKeySpec);
			}
			else if(ng.isDHE()) {
				publicKeyY = new BigInteger(1,remoteKeyShare.keyExchange);
				DHPublicKeySpec dhKeySpec = new DHPublicKeySpec(publicKeyY, ((DHPrivateKey)localDHPrivateKey).getParams().getP(), ((DHPrivateKey)localDHPrivateKey).getParams().getG());
				KeyFactory localKeyFactory = KeyFactory.getInstance("DiffieHellman");
				remoteDHPublicKey = (DHPublicKey) localKeyFactory.generatePublic(dhKeySpec);
			}

			if(ng.isECDHE())
				keyAgree = KeyAgreement.getInstance("ECDH");
			else if(remoteKeyShare.getGroup().isDHE())
				keyAgree = KeyAgreement.getInstance("DiffieHellman");
			
			keyAgree.init(localDHPrivateKey);
			keyAgree.doPhase(remoteDHPublicKey, true);
			DHsecret = keyAgree.generateSecret();
		} catch (Exception e) {
			DebugInfo.printMessage(1, "!!! Error: TLSHandshakeParameters::createDHEsecret: failed ");
			e.printStackTrace();
		}

		DebugInfo.printDumpWithLength(1, DHsecret, "---DHsecret: ");
		return DHsecret;
	}

	public void printDump(int debugLevel) {
		DebugInfo.printMessage(debugLevel, "TLSHandshakeParameters:");
		DebugInfo.printMessage(debugLevel, "   isTLS13: " + isTLS13);
		DebugInfo.printMessage(debugLevel, "   keyExchangeMode: " + keyExchangeMode);
		DebugInfo.printMessage(debugLevel, "\n");
	}
}
