package ru.ispras.telecomtesk.tlsv13.TLSState;

import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.SecureRandom;
import java.security.spec.ECGenParameterSpec;
import java.util.ArrayList;

import javax.crypto.Mac;
import javax.crypto.spec.DHParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.HandshakeMessages.HandshakeMessage;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.HandshakeMessageType;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.MACAlgorithm;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.NamedGroup;

public class TLSKeyDerivation
{
	public static int debugLevel = 7;
	public static int debugLevelTest = 1;
	
	public static KeyPair generateKeyPair(NamedGroup group) throws Exception {
		KeyPairGenerator keyGen = null;
		if(group.isDHE()) {
			DHParameterSpec dhspec = new DHParameterSpec(group.getDHEmodulus(), group.getDHEgenerator());
			keyGen = KeyPairGenerator.getInstance("DiffieHellman");
        	keyGen.initialize(dhspec);
		} else if(group.isECDHE()) {
		    keyGen = KeyPairGenerator.getInstance("EC");
		    keyGen.initialize(new ECGenParameterSpec(group.getECname()), new SecureRandom());
		}

		return keyGen.generateKeyPair();
	}
	
	public static byte[] hash(MACAlgorithm alg, byte[] data) throws Exception {
		byte[][] dataArray = {data};
		return hash(alg, dataArray);
	}
	
	public static byte[] hash(MACAlgorithm alg, byte[][] data) throws Exception {
		MessageDigest mobj;
		byte[] res = null;
		String algName;
		
		if(alg == MACAlgorithm.mac_null)
			return new byte[0];
		
		DebugInfo.printDump(TLSKeyDerivation.debugLevel,data,"TLSKeyDerivation::hash:");
		
		algName = alg.getHashAlgorithmName();
		mobj = MessageDigest.getInstance(algName);
		if(data != null) {
			for(int i=0; i<data.length; i++) {
				if(data[i] != null)
					mobj.update(data[i]);
			}
		}
		res = mobj.digest();
		
		return res;
	}
	
	public static byte[] mac(MACAlgorithm alg, byte[] key, byte[] data) throws Exception {
		byte[][] dataArray = {data};
		return mac(alg, key, dataArray);
	}
	
	public static byte[] mac(MACAlgorithm alg, byte[] key, byte[][] data) throws Exception {
		Mac mobj;
		byte[] res = null;
		SecretKeySpec sk;
		String algName;
		
		if(alg == MACAlgorithm.mac_null)
			return new byte[0];
		
		if(key != null) {
			algName = alg.getMacAlgorithmName();
			mobj = Mac.getInstance(algName);
			sk = new SecretKeySpec(key, algName);
			mobj.init(sk);
			if(data != null) {
				for(int i=0; i<data.length; i++) {
					if(data[i] != null)
						mobj.update(data[i]);
				}
			}
			res = mobj.doFinal();
		}
		
		return res;
	}
	
	public static byte[] TLS_HKDF_Extract (MACAlgorithm alg, byte[] salt, byte[] ikm) throws Exception {
		if(salt == null) {
			salt = new byte[alg.getHashLength()];
			for(int i=0; i<salt.length; i++) 
				salt[i] = 0;
		}
		if(ikm == null) {
			ikm = new byte[alg.getHashLength()];
			for(int i=0; i<ikm.length; i++) 
				ikm[i] = 0;
		}

		DebugInfo.printDump(TLSKeyDerivation.debugLevel, salt, "TLSKeyDerivation::TLS_HKDF_Extract:salt: " + alg);
		DebugInfo.printDump(TLSKeyDerivation.debugLevel, ikm, "TLSKeyDerivation::TLS_HKDF_Extract:ikm: ");
		return TLSKeyDerivation.mac(alg, salt, ikm);
	}

	public static byte[] TLS_HKDF_Expand (MACAlgorithm alg, byte[] prk, byte[] info, int length) throws Exception {
		byte[] res = null;
		int n, pos = 0, curlen;
		int hashlen = alg.getHashLength();
		byte[] curT = null;
		byte[][] curInfo = new byte[3][];

		if((hashlen > 0) && (length > 0)) {
			res = new byte[length];
			n = (int) Math.ceil((double)length / (double)hashlen);
			curInfo[1] = info;
			curInfo[2] = new byte[1];

			for(int i=1; i<=n; i++) {
				curInfo[0] = curT;
				curInfo[2][0] = (byte)i;
				curT = TLSKeyDerivation.mac(alg, prk, curInfo);
				curlen = (length >= pos + curT.length) ? curT.length : (length - pos);
				System.arraycopy(curT, 0, res, pos, curlen);
				pos += curlen;
			}
		}

		DebugInfo.printDump(TLSKeyDerivation.debugLevel, res, "TLSKeyDerivation::TLS_HKDF_Expand:res ");
		return res;
	}

	public static byte[] TLS_HKDF_ExpandLabel (MACAlgorithm alg, byte[] key, String label, byte[] context) throws Exception {
		return TLS_HKDF_ExpandLabel(alg,key,label,context,alg.getHashLength());
	}
	
	public static byte[] TLS_HKDF_ExpandLabel (MACAlgorithm alg, byte[] key, String label, byte[] context, int length) throws Exception {
		int pos = 0, hkdfLabelLen = 0;
		byte[] labelRaw = ("tls13 " + label).getBytes();
		
		hkdfLabelLen = 4 + labelRaw.length;
		if(context != null) hkdfLabelLen += context.length;

		byte[] hkdfLabel = new byte[hkdfLabelLen];
		hkdfLabel[pos++] = (byte)(length >>> 8);
		hkdfLabel[pos++] = (byte)length;

		hkdfLabel[pos++] = (byte)labelRaw.length;
		System.arraycopy(labelRaw, 0, hkdfLabel, pos, labelRaw.length);
		pos += labelRaw.length;
		if(context != null) { 
			hkdfLabel[pos++] = (byte)context.length;
			System.arraycopy(context, 0, hkdfLabel, pos, context.length);
		} else
			hkdfLabel[pos++] = 0;

		DebugInfo.printDump(TLSKeyDerivation.debugLevel, hkdfLabel, "TLSKeyDerivation::TLS_HKDF_ExpandLabel:hkdfLabel: ");
		return TLSKeyDerivation.TLS_HKDF_Expand(alg, key, hkdfLabel, length);
	}

	public static byte[] TLS_DeriveSecret (MACAlgorithm alg, byte[] key, String label, byte[] messagesHash) throws Exception {
		DebugInfo.printMessage(TLSKeyDerivation.debugLevel, "TLSKeyDerivation::TLS_DeriveSecret: ");
		return TLSKeyDerivation.TLS_HKDF_ExpandLabel(alg, key, label, messagesHash);
	}

	public static byte[] TLS_DeriveSecret (MACAlgorithm alg, byte[] key, String label) throws Exception {
		DebugInfo.printMessage(TLSKeyDerivation.debugLevel, "TLSKeyDerivation::TLS_DeriveSecret: ");
		return TLSKeyDerivation.TLS_HKDF_ExpandLabel(alg, key, label, TLS_TranscriptHash(alg, (ArrayList<HandshakeMessage>) null));
	}

	public static byte[] TLS_DeriveSecret (MACAlgorithm alg, byte[] key, String label, ArrayList<HandshakeMessage> messages) throws Exception {
		DebugInfo.printMessage(TLSKeyDerivation.debugLevel, "TLSKeyDerivation::TLS_DeriveSecret: ");
		return TLSKeyDerivation.TLS_HKDF_ExpandLabel(alg, key, label, TLS_TranscriptHash(alg, messages));
	}

	public static byte[] TLS_TranscriptHash (MACAlgorithm alg, ArrayList<HandshakeMessage> messages) throws Exception {
		byte[][] msgHash = null;
		int i;
		if (messages != null) {
			msgHash = new byte[messages.size()][];
			for(i=0; i < messages.size(); i++) {
				msgHash[i] = messages.get(i).net_msg;
			}
			DebugInfo.printDump(TLSKeyDerivation.debugLevel, msgHash, "TLSKeyDerivation::TLS_TranscriptHash: " + i + " handshake messages");
		} else {
			DebugInfo.printDump(TLSKeyDerivation.debugLevel, msgHash, "TLSKeyDerivation::TLS_TranscriptHash: no messages");
//			return null;
		}
		return TLSKeyDerivation.hash(alg, msgHash);
	}

	public static TLStrafficKey TLS_TrafficKey (MACAlgorithm alg, byte[] secret, int keyLength, int ivLength) throws Exception {
		return new TLStrafficKey(
					TLS_HKDF_ExpandLabel(alg, secret, "key", null, keyLength),
					TLS_HKDF_ExpandLabel(alg, secret, "iv", null, ivLength)
					);
	}

	//  application_traffic_secret_N+1 = HKDF-Expand-Label(application_traffic_secret_N, "traffic upd", "", Hash.length)

	public static byte[] TLS_UpdateApplicationKey (MACAlgorithm alg, byte[] secret_n) throws Exception {
		return TLS_HKDF_ExpandLabel(alg, secret_n, "traffic upd", null);
	}
	
}
