package ru.ispras.telecomtesk.tlsv13.TLSState;

import java.util.Random;

import javax.crypto.Cipher;
import javax.crypto.Mac;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;

import ru.ispras.telecomtesk.tlsv13.ConvertTypes;
import ru.ispras.telecomtesk.tlsv13.DebugInfo;
import ru.ispras.telecomtesk.tlsv13.TLSAlgorithms.*;

public class TLSSecurityParameters
{
	public TLSCipherSuite cs = TLSCipherSuite.TLS_NULL_WITH_NULL_NULL;
	public KeyExchangeAlgorithm ke_alg = KeyExchangeAlgorithm.NULL;
	public PRFAlgorithm prf = PRFAlgorithm.tls_prf_sha256;
	public BulkCipherAlgorithm bulk_cipher_alg = BulkCipherAlgorithm.bulk_null;
	public CipherType cipher_type = CipherType.stream;
	public byte enc_key_length = 0;
	public byte block_length = 0;
	public byte fixed_iv_length = 0;
	public byte record_iv_length = 0;
	public MACAlgorithm mac_alg = MACAlgorithm.mac_null;
	public byte mac_length = 0;
	public byte mac_key_length = 0;
	public byte compression_alg = 0;
	public byte[] presetAAD = null;
	public byte[] client_random = null;
	public byte[] server_random = null;
	public byte[] psk = null;
	public byte[] early_secret = null;
	public byte[] binder_key = null;
	public byte[] client_early_traffic_secret = null;
	public byte[] early_exporter_master_secret = null;
	public byte[] dheSecret = null;
	public byte[] handshake_secret = null;
	public byte[] client_handshake_traffic_secret = null;
	public byte[] server_handshake_traffic_secret = null;
	public byte[] master_secret = null;
	public byte[] client_application_traffic_secret = null;
	public byte[] server_application_traffic_secret = null;
	public byte[] exporter_master_secret = null;
	public byte[] resumption_master_secret = null;
	public TLStrafficKey clientKey;
	public TLStrafficKey serverKey;
	public enum PaddingFillStyle {FILL_PAD_ZERO, FILL_PAD_NUM};
	public PaddingFillStyle paddingStyle = PaddingFillStyle.FILL_PAD_ZERO;
	public int paddingLength = 0;
	
	public TLSSecurityParameters() {
		clientKey = new TLStrafficKey();
		serverKey = new TLStrafficKey();
                master_secret = new byte[48]; client_random = new byte[32]; server_random = new byte[32];
	}

	public boolean isUseCrypto() {
		return (bulk_cipher_alg != BulkCipherAlgorithm.bulk_null); 
	}

	public TLStrafficKey getReadKey(ConnectionRole localSide) {
		if(localSide == ConnectionRole.CLIENT) return serverKey;
		else return clientKey;
	}
	
	public TLStrafficKey getWriteKey(ConnectionRole localSide) {
		if(localSide == ConnectionRole.CLIENT) return clientKey;
		else return serverKey;
	}
	
	/*
	 *      master_secret = PRF(pre_master_secret, "master secret",
     *                    ClientHello.random + ServerHello.random)
     *                    [0..47];
     *      When Diffie-Hellman is used for server authentication and key exchange, 
     *      the negotiated key (Z) is used as the pre_master_secret.
     *      Leading bytes of Z that contain all zero bits are stripped before it is used.
 	 */
	public int createMasterSecret(byte[] pre_master_secret) throws Exception {
		int res = 0, i;
		byte[] rnd = new byte[64];
		byte[] new_pre_master_secret = pre_master_secret;
		boolean isDHalgorithm = KeyExchangeAlgorithm.isDHalgorithm(ke_alg);
		
		if((pre_master_secret != null)&&(pre_master_secret.length != 0)) {
			if(isDHalgorithm == true) {
				for(i=0;i<pre_master_secret.length;i++)
					if(pre_master_secret[i] != 0) break;
				
				if(i != 0) {
					new_pre_master_secret = new byte[pre_master_secret.length - i];
					System.arraycopy(pre_master_secret, i, new_pre_master_secret, 0, pre_master_secret.length - i);
				}
			}
			System.arraycopy(client_random, 0, rnd, 0, 32);
			System.arraycopy(server_random, 0, rnd, 32, 32);
			res = prf(new_pre_master_secret, "master secret", rnd, master_secret);
		}
		else
			res = 10;
		
		return res;
	}
	
	public void setMasterSecret(byte[] ms){
		if(ms != null) {
			master_secret = new byte[ms.length];
			System.arraycopy(ms, 0, master_secret, 0, ms.length);
		}
	}

	public byte[] getMasterSecret(){
		return master_secret;
	}

	public void setRandom(byte[] rnd, boolean isClientRandom){
		if((rnd == null) || (rnd.length != 32))
			System.out.println("TLSSecurityParameters::setRandom: byte[] = NULL or byte[].length != 32");
		else {
			if(isClientRandom == true)
				System.arraycopy(rnd, 0, client_random, 0, 32);
			else
				System.arraycopy(rnd, 0, server_random, 0, 32);
		}
	}

	/*
	 *       P_hash(secret, seed) = HMAC_hash(secret, A(1) + seed) +
     *                        		HMAC_hash(secret, A(2) + seed) +
     *                       		HMAC_hash(secret, A(3) + seed) + ... 
     *                       
     *                       where  A(0) = seed
     *								A(i) = HMAC_hash(secret, A(i-1))
	 */
	public int prf(byte[] secret, String label, byte[] seed, byte[] target) throws Exception {
		int res=0, curpos=0, len, maclen;
		String alg;
		Mac mobj;
		SecretKeySpec sk;
		byte[] curSeed;
		byte[] curHash;
		byte[] lseed;
		
		if((target != null)&&(secret != null)&&(seed != null)) {
			if(target.length == 0) return 0;
			
				switch (prf) {
					case tls_prf_sha256:
						alg = "HmacSHA256";
						break;
					default:
						System.out.println("TLSSecurityParameters::prf: Algorithm not supported");
						return -2;
				}
				
				if(label != null)
					lseed = label.getBytes("ASCII");
				else 
					lseed = new byte[0];
				
				mobj = Mac.getInstance(alg);
				sk = new SecretKeySpec(secret, alg);
				maclen = mobj.getMacLength();
				curSeed = new byte[maclen];
				curHash = new byte[maclen];
				mobj.init(sk);
				// create A(1)
				mobj.update(lseed);
				mobj.update(seed);
				mobj.doFinal(curSeed,0);
				
				// while (curpos != target.length)
				while (true) {
					mobj.update(curSeed);
					mobj.update(lseed);
					mobj.update(seed);
					mobj.doFinal(curHash,0);
					
					len = (maclen <= target.length - curpos) ? maclen : target.length - curpos;
					System.arraycopy(curHash, 0, target, curpos, len);
					curpos += len;
					if(curpos == target.length) break;
					
					// create next A(i)
					mobj.update(curSeed);
					mobj.doFinal(curSeed,0);
				}
		}
		else
			res = -1;
		
		return res;
	}

	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(int data) {
		int res = -1;
		
		if((data >= 0) && (cipher_type == CipherType.block)) {
			if(block_length != 0) 
				res = block_length - data % block_length;
		}
		return res;
	}

	// include one byte for 'Padding_length' field
	public int getMaxPaddingSize(int minPaddingSize) {
		int res = -1;
		
		if(minPaddingSize <= 256)
			res = 256 - ((256 - minPaddingSize) % block_length);
		
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(byte[] data) {
		int res = -1;
		
		if((data != null) && (cipher_type == CipherType.block)) {
			if(block_length != 0) 
				res = block_length - data.length % block_length;
		}
		return res;
	}
	
	// include one byte for 'Padding_length' field
	public int getMinPaddingSize(byte[][] data) {
		int res = -1;
		int datalen = 0;
		
		if((data != null) && (cipher_type == CipherType.block)) {
			if(block_length != 0) {
				for(int i=0; i<data.length; i++){
					if(data[i] != null)
						datalen += data[i].length;
				}
				if(datalen != 0)
					res = block_length - datalen % block_length;
			}
		}
		return res;
	}

	public void setAlgorithmParams() {
		setAlgorithmParams(this.cs);
	}

	public void setAlgorithmParams(TLSCipherSuite cs) {
		this.cs = cs;
		if(cs == TLSCipherSuite.UNKNOWN)
			return;
		
		switch(cs) {
		case TLS_NULL_WITH_NULL_NULL:
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.mac_null; break;
		case TLS_RSA_WITH_NULL_MD5:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.md5; break;
		case TLS_RSA_WITH_NULL_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.sha1; break;
		case TLS_RSA_WITH_NULL_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_null; mac_alg = MACAlgorithm.sha256; break;
		case TLS_RSA_WITH_RC4_128_MD5:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.md5; break;
		case TLS_RSA_WITH_RC4_128_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.sha1; break;
		case TLS_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;
		case TLS_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;
		case TLS_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;
		case TLS_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break;
		case TLS_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break;
		
		case TLS_DH_DSS_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;
		case TLS_DH_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DHE_DSS_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;     
		case TLS_DHE_RSA_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;     
		case TLS_DH_anon_WITH_RC4_128_MD5:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_rc4; mac_alg = MACAlgorithm.md5; break;          
		case TLS_DH_anon_WITH_3DES_EDE_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_3des; mac_alg = MACAlgorithm.sha1; break;  
		case TLS_DH_DSS_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;       
		case TLS_DH_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;       
		case TLS_DHE_DSS_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DHE_RSA_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DH_anon_WITH_AES_128_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DH_DSS_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;       
		case TLS_DH_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;       
		case TLS_DHE_DSS_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DHE_RSA_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;      
		case TLS_DH_anon_WITH_AES_256_CBC_SHA:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha1; break;   
		case TLS_DH_DSS_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break; 
		case TLS_DH_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break; 
		case TLS_DHE_DSS_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break;   
		case TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break;   
		case TLS_DH_anon_WITH_AES_128_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128; mac_alg = MACAlgorithm.sha256; break; 
		case TLS_DH_DSS_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break;    
		case TLS_DH_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break;    
		case TLS_DHE_DSS_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_DSS;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break;
		case TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DHE_RSA;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break;
		case TLS_DH_anon_WITH_AES_256_CBC_SHA256:
			ke_alg = KeyExchangeAlgorithm.DH_ANON;
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256; mac_alg = MACAlgorithm.sha256; break; 

			// tls v1.3
		case TLS_AES_128_GCM_SHA256:
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128_gcm; mac_alg = MACAlgorithm.sha256; break;   
		case TLS_AES_256_GCM_SHA384:
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes256_gcm; mac_alg = MACAlgorithm.sha384; break;   
		case TLS_AES_128_CCM_SHA256:
			bulk_cipher_alg = BulkCipherAlgorithm.bulk_aes128_ccm; mac_alg = MACAlgorithm.sha256; break;   

		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: TLSCipherSuite not supported");
		}
		
		switch(bulk_cipher_alg) {
		case bulk_null:
			cipher_type = CipherType.stream;
			enc_key_length = 0; block_length = 0; record_iv_length = 0; break;
		case bulk_rc4:
			cipher_type = CipherType.stream;
			enc_key_length = 16; block_length = 0; record_iv_length = 0; break;
		case bulk_3des:
			cipher_type = CipherType.block;
			enc_key_length = 24; block_length = 8; record_iv_length = 8; break;
		case bulk_aes128:
			cipher_type = CipherType.block;
			enc_key_length = 16; block_length = 16; record_iv_length = 16; break;
		case bulk_aes128_ccm:
		case bulk_aes128_gcm:
			cipher_type = CipherType.aead;
			enc_key_length = 16; block_length = 16; record_iv_length = 12; break;
		case bulk_aes256:
		case bulk_aes256_gcm:
			cipher_type = CipherType.aead;
			enc_key_length = 32; block_length = 16; record_iv_length = 12; break;
		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: Encrypt algorithm not supported");
		}
		
		switch(mac_alg) {
		case mac_null:
			mac_length = 0; mac_key_length = 0; break;
		case md5:
			mac_length = 16; mac_key_length = 16; break;
		case sha1:
			mac_length = 20; mac_key_length = 20; break;
		case sha256:
			mac_length = 32; mac_key_length = 32; break;
		case sha384:
			mac_length = 48; mac_key_length = 48; break;
		default:
			System.out.println("TLSSecurityParameters::setAlgorithmParams: MAC algorithm not supported");
		}
	}

	public byte[] createPadding(int padLen) {
		paddingLength = padLen;
		return createPadding();
	}
	
	public byte[] createPadding() {
		byte[] res = new byte[paddingLength];

		for(int i=0;i<res.length;i++){
			switch (paddingStyle) {
				case FILL_PAD_ZERO:
					res[i] = 0; break;
				case FILL_PAD_NUM:
					res[i] = (byte)(i+1); break;
			}
		}

		return res;
	}
	
	public static boolean checkPadding(byte[] data) {
		boolean res = false;
		if(data != null) {
			int len = data.length - 1;
			for(int i=0;i<=len;i++){
				if(data[i] != 0) 
					return false;
			}
			res = true;
		}
		return res;
	}
	
	public static boolean checkPadding(byte[] data, int offset) {
		boolean res = false;
		if(data != null && offset < data.length) {
			int len = data.length - 1 - offset;
			for(int i=0;i<=len;i++){
				if(data[offset + i] != 0) 
					return false;
			}
			res = true;
		}
		return res;
	}
	
	public byte[] encrypt(byte[] data, ConnectionRole localSide) throws Exception {
		byte[] res = null;
		SecretKeySpec sk;
		String alg, algmode;
		Cipher cobj;
		TLStrafficKey writeKey;
		
		DebugInfo.printMessage(4,"TLSSecurityParameters::encrypt: bulk_cipher_alg: " + bulk_cipher_alg);
		
		switch (bulk_cipher_alg) {
			case bulk_null:
				return data;
			case bulk_aes128_ccm:
				alg = "AES";
				algmode = "AES/CCM/NoPadding"; break;
			case bulk_aes128_gcm:
			case bulk_aes256_gcm:
				alg = "AES";
				algmode = "AES/GCM/NoPadding"; break;
			default:
				DebugInfo.printMessage(2, "TLSSecurityParameters::encrypt: Algorithm not supported");
				return null;
		}
		
		writeKey = getWriteKey(localSide);
		DebugInfo.printMessage(4,"writeKey Sequence number " + writeKey.sec_num);
		DebugInfo.printDump(4, writeKey.key, "writeKey: ");

		sk = new SecretKeySpec(writeKey.key, alg);
		byte[] iv = writeKey.getNonce();
		cobj = Cipher.getInstance(algmode);
		GCMParameterSpec gcmSpec = new GCMParameterSpec(8 * cobj.getBlockSize(), iv);
//		IvParameterSpec ivSpec = new IvParameterSpec(iv);
		cobj.init(Cipher.ENCRYPT_MODE, sk, gcmSpec);

		// encrypt data
		int enclen = cobj.getOutputSize(data.length);
		cobj.updateAAD(getDefaultGCMAAD(enclen));
		DebugInfo.printDump(4, getDefaultGCMAAD(enclen), "TLSSecurityParameters::encrypt: before encryption: GCMAAD");
		DebugInfo.printDump(4, iv, "TLSSecurityParameters::encrypt: before encryption: iv");
		DebugInfo.printDump(4, data, "TLSSecurityParameters::encrypt: before encryption: data");
		res = cobj.doFinal(data); 
		DebugInfo.printDump(4, res, "TLSSecurityParameters::encrypt: after encryption: data");
		DebugInfo.printMessage(1, "TLSSecurityParameters::encrypt: encrypted length for GCM AAD " + enclen);
		if(enclen != res.length)
			DebugInfo.printMessage(1, "===Inner Error!GCM AAD length wrong");
	
		return res;
	}
	
	public byte[] encrypt(byte[][] data, ConnectionRole localSide) throws Exception {
		return encrypt(ConvertTypes.combineArrays(data), localSide);
	}

	public byte[] decrypt(byte[] data, ConnectionRole localSide) throws Exception {
		byte[] res = null;
		SecretKeySpec sk;
		String alg, algmode;
		Cipher cobj;
		TLStrafficKey readKey;
		
		DebugInfo.printDump(6, data, "TLSSecurityParameters::decrypt: encrypted data: ");
		DebugInfo.printMessage(4,"TLSSecurityParameters::decrypt: bulk_cipher_alg: " + bulk_cipher_alg);
		
		switch (bulk_cipher_alg) {
			case bulk_null:
				return data;
			case bulk_aes128_ccm:
				alg = "AES";
				algmode = "AES/CCM/NoPadding"; break;
			case bulk_aes128_gcm:
			case bulk_aes256_gcm:
				alg = "AES";
				algmode = "AES/GCM/NoPadding"; break;
			default:
				DebugInfo.printMessage(2, "TLSSecurityParameters::decrypt: Algorithm not supported");
				return null;
		}
		
		readKey = getReadKey(localSide);
		DebugInfo.printMessage(4,"readKey Sequence number " + readKey.sec_num);
		DebugInfo.printDump(4, readKey.key, "readKey: ");
		
		sk = new SecretKeySpec(readKey.key, alg);
		byte[] iv = readKey.getNonce();
		cobj = Cipher.getInstance(algmode);
		GCMParameterSpec gcmSpec = new GCMParameterSpec(8 * cobj.getBlockSize(), iv);
//		IvParameterSpec ivSpec = new IvParameterSpec(iv);
		cobj.init(Cipher.DECRYPT_MODE, sk, gcmSpec);
		cobj.updateAAD(getDefaultGCMAAD(data.length));
		DebugInfo.printDump(4, getDefaultGCMAAD(data.length), "TLSSecurityParameters::decrypt: before decryption: GCMAAD");
		DebugInfo.printDump(4, iv, "TLSSecurityParameters::decrypt: before decryption: iv");
		DebugInfo.printDump(4, data, "TLSSecurityParameters::decrypt: before decryption: data");
		res = cobj.doFinal(data); 
		
		return res;
	}
	
	public byte[] getDefaultGCMAAD(int datalen) {
		// ContentType opaque_type = application_data = 23
		// ProtocolVersion legacy_record_version = 0x0303
		// uint16 length
		if(presetAAD == null) {
			byte[] aad = {23,3,3,(byte)(datalen >>> 8),(byte)datalen};
			return aad;
		}
		else
			return presetAAD;
	}

	public TLStrafficKey deriveTrafficKey (byte[] key) throws Exception {
		return TLSKeyDerivation.TLS_TrafficKey(this.mac_alg, key, this.enc_key_length, this.record_iv_length);
	}
	
	public void createHandshakeTrafficKeys() throws Exception {
		clientKey = deriveTrafficKey(client_handshake_traffic_secret);
		serverKey = deriveTrafficKey(server_handshake_traffic_secret);
	}
	
	public void createClientApplicationTrafficKeys() throws Exception {
		clientKey = deriveTrafficKey(client_application_traffic_secret);
	}
	
	public void createServerApplicationTrafficKeys() throws Exception {
		serverKey = deriveTrafficKey(server_application_traffic_secret);
	}
	
	public byte[] updateApplicationKey(byte[] key) throws Exception {
		return TLSKeyDerivation.TLS_UpdateApplicationKey(this.mac_alg, key);
	}
	
	//  application_traffic_secret_N+1 = HKDF-Expand-Label(application_traffic_secret_N, "traffic upd", "", Hash.length)
	public void updateClientTrafficKeys () throws Exception {
		client_application_traffic_secret = updateApplicationKey(client_application_traffic_secret);
		createClientApplicationTrafficKeys();
	}
	
	public void updateServerTrafficKeys () throws Exception {
		server_application_traffic_secret = updateApplicationKey(server_application_traffic_secret);
		createServerApplicationTrafficKeys();
	}
	
	public void printDump(int debugLevel) {
		DebugInfo.printMessage(debugLevel, "TLSSecurityParameters:");
		DebugInfo.printMessage(debugLevel, "   TLSCipherSuite: " + cs.name());
		DebugInfo.printMessage(debugLevel, "   BulkCipherAlgorithm: " + bulk_cipher_alg.name());
		DebugInfo.printMessage(debugLevel, "   enc_key_length: " + enc_key_length);
		DebugInfo.printMessage(debugLevel, "   record_iv_length: " + record_iv_length);
		DebugInfo.printMessage(debugLevel, "   MACAlgorithm: " + mac_alg.name());
		DebugInfo.printMessage(debugLevel, "   mac_length: " + mac_length);
		DebugInfo.printMessage(debugLevel, "   mac_key_length: " + mac_key_length);
		DebugInfo.printDump(debugLevel, client_random, "   client_random: ");
		DebugInfo.printDump(debugLevel, server_random, "   server_random: ");
		DebugInfo.printDump(debugLevel, psk, "   psk: ");
		DebugInfo.printDump(debugLevel, early_secret, "   early_secret: ");
		DebugInfo.printDump(debugLevel, binder_key, "   binder_key: ");
		DebugInfo.printDump(debugLevel, client_early_traffic_secret, "   client_early_traffic_secret: ");
		DebugInfo.printDump(debugLevel, early_exporter_master_secret, "   early_exporter_master_secret: ");
		DebugInfo.printDump(debugLevel, dheSecret, "   dheSecret: ");
		DebugInfo.printDump(debugLevel, handshake_secret, "   handshake_secret: ");
		DebugInfo.printDump(debugLevel, client_handshake_traffic_secret, "   client_handshake_traffic_secret: ");
		DebugInfo.printDump(debugLevel, server_handshake_traffic_secret, "   server_handshake_traffic_secret: ");
		DebugInfo.printDump(debugLevel, master_secret, "   master_secret: ");
		DebugInfo.printDump(debugLevel, client_application_traffic_secret, "   client_application_traffic_secret_n: ");
		DebugInfo.printDump(debugLevel, server_application_traffic_secret, "   server_application_traffic_secret_n: ");
		DebugInfo.printDump(debugLevel, exporter_master_secret, "   exporter_master_secret: ");
		DebugInfo.printDump(debugLevel, resumption_master_secret, "   resumption_master_secret: ");
		DebugInfo.printMessage(debugLevel,"clientKey Sequence number " + clientKey.sec_num);
		DebugInfo.printDump(debugLevel, clientKey.key, "clientKey: ");
		DebugInfo.printDump(debugLevel, clientKey.iv, "clientKey iv: ");
		DebugInfo.printMessage(debugLevel,"serverKey Sequence number " + serverKey.sec_num);
		DebugInfo.printDump(debugLevel, serverKey.key, "serverKey: ");
		DebugInfo.printDump(debugLevel, serverKey.iv, "serverKey iv: ");
		DebugInfo.printMessage(debugLevel, "   padding: " + paddingStyle + " padLen " + paddingLength);
		DebugInfo.printMessage(debugLevel, "\n");
	}

	public void printTLStrafficKeys(int debugLevel) {
		DebugInfo.printMessage(debugLevel,"current TLS traffic keys:");
		DebugInfo.printMessage(debugLevel,"clientKey Sequence number " + clientKey.sec_num);
		DebugInfo.printDump(debugLevel, clientKey.key, "clientKey: ");
		DebugInfo.printDump(debugLevel, clientKey.iv, "clientiv: ");
		DebugInfo.printMessage(debugLevel,"serverKey Sequence number " + serverKey.sec_num);
		DebugInfo.printDump(debugLevel, serverKey.key, "serverKey: ");
		DebugInfo.printDump(debugLevel, serverKey.iv, "serveriv: ");
	}
	
}
